//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalView.h	--- several classes defining a CalView object
//
//	DESCRIPTION:
//		Several classes are declared which define a CalViewServer object.
//		There is a single global instance of a CalViewServer object which
//		applications (i.e. calview) should use. The classes and their
//		methods defined	herein have been designed with the aim of
//		providing maximum fidelity and usefulness for a GUI-based
//		calview application.
//
//	CHANGE HISTORY:
//		UL, 1999-03-24: 1.0.0: - skeleton
//		UL, 1999-??-??: 1.1.0: - first complete version
//		UL, 1999-08-12: 1.2.0: - CalViewServer::viewable() now builds
//								 up a new vector every time it is called
//								 which just the viewables that are applicable
//								 to the current instrument as set in the
//								 Cal state
//							   - added CalViewable::isApplicable()
//		UL, 2000-03-24: 1.3.0: - added state variable 'spectralOrder'
//
//	AUTHOR:
//		U. Lammers, Wed Mar 24 14:29:44 MET 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/calview/CalView.h,v 1.13 2002/05/24 10:59:16 gvacanti Exp $
//

#ifndef CALVIEW_H
#define CALVIEW_H

#include "StateVariable.h"	//	general Cal include file
#include "CalChild.h"		//	general Cal include file
#include "Xmm.h"			//	instrument IDs


//
//	CalViewState: define the state of the CalView object by a set of
//				  variables
class CalViewState {
public:
	CalViewState() : E(100.,	//	default energy is 0 eV
					   "E",		//	name of variable
					   "eV",	//	we do things in eV
					   "energy",//	What is it?
					   true,	//	there is a lower limit
					   true,	//	there is an upper limit
					   100.,	//	no energies lower than this
					   15000.),	//	no energies greater than 15 keV allowed
					 theta(0.,	//	on-axis
					   "theta",	//	name of variable
					   "arcsec",//	unit
					   "off-axis angle",//	what is it
					   true,	//	there is a lower limit
					   true,	//	there is an upper limit
					   0.,		//	no negative off-axis angles allowed
					   1800.),	//	no off-axis angles > .5deg allowed
					 phi(0.,	//	azimuth
					   "phi",	//	name of variable
					   "deg",	//	unit
					   "azimuth angle",//	what is it
					   true,	//	there is a lower limit
					   true,	//	there is an upper limit
					   -360.,	//	no smaller angles allowed
					   360.),	//	no greater angles allowed
					spectralOrder(0,
					   "m",		//	name of variable
					   "",		//	no units
					   "spectral order", // what is it
					   true,	//	has min
					   true,	//	has max
					   -5,		//	min allowed order
					   0)		//	max allowed order
					

		{}
		//
		//	PURPOSE:
		//		Create a new CalViewState object
		//

	RangeStateVariable<double>	E,
								theta,
								phi;
	RangeStateVariable<int>		spectralOrder;
		//
		//	This is the list of variables defining the state:
		//		E		: energy for plots of redistribution, PSF, etc.
		//		theta	: off-axis angle (Ae) [arcsec]
		//		phi		: azimuth angle (Ae)  [deg]
		//		spectralOrder:
		//				  spectral order for RGS [unitless]
		//


	~CalViewState() {}
		//
		//	PURPOSE:
		//		destroy object constructed with above constructor
}; // CalViewState


//
//	CalViewProductDescriptor:
//		utility class to describe the type and location
//		of the products created by the CalViewable::view methods below
//
class CalViewProductDescriptor {
public:
	enum Type { XMGR_PARM, FITS_IMG, ASCII };
		//
		//	XMGR_PARM: an XMGR parameter file; xmgr can be directly invoked
		//			   on this to show the plot
		//	FITS_IMG : a >=2D FITS image to be displayed with a FITS image
		//			   viewer (SAOtng, SAOds9, etc.)
		//	ASCII	 : ASCII file to be displayed in editor, text widget, etc.
		//

	CalViewProductDescriptor(const std::string &location,
							 CalViewProductDescriptor::Type type = XMGR_PARM)
		: _location(location), _type(type) {}
		//
		//	PURPOSE:
		//		Construct a CalViewProductDescriptor object
		//
		//	ARGUMENTS:
		//		In:
		//			location: file name (incl. path) of generated product
		//			type	: type of product
		//

	Type type() const { return _type; }
	const std::string &location() const { return _location; }
		//
		//	PURPOSE:
		//		Inquire type and location of an CalViewProduct

	~CalViewProductDescriptor() {}
		//
		//	PURPOSE:
		//		destroy an object constructed with above constructor
		//

private:
	std::string	_location;
	Type		_type;
}; // CalViewProduct


class CalViewServer;	//	forward declaration


//
//	CalViewable:
//		Objects of class CalViewable represent entities which possess
//		a view() method to visualize a particular calibration quantity.
//		CalViewables can contain subsidiary CalViewables which can
//		visualize the calibration quantity in question in a different
//		form or context. CalViewables() inquire the state of the
//		CalViewServer.
//
class CalViewable : public CalChild<CalViewServer> {
public:
	CalViewable(const CalViewServer *server = 0);
		//
		//	PURPOSE:
		//		Construct a CalViewable object
		//

	const std::string &name() const;
		//
		//	PURPOSE:
		//		Return a unique single-word name of the Viewable - can
		//		be used as a label in a menu entry

	const std::string &whatIs() const;
		//
		//	PURPOSE:
		//		Return a one-line description of the Viewable - can
		//		be used as text in a Tooltip associated with
		//		a menu entry for this Viewable
		//


	virtual bool isApplicable(Xmm::Instrument instrument) const;
		//
		//	RETURNS:
		//		True or false depending on whether the Viewable makes
		//		sense for the specified instrument. If the Viewable
		//		is not applicable, calling the view method will lead
		//		to a an error condition
		//


	virtual CalViewProductDescriptor view(const std::string &setName = "")
	const;
		//
		//	PURPOSE:
		//		Create a data product which represents a visual
		//		representation of the Viewable object, e.g. a 1D plot
		//		or an image. The type and location of the product
		//		is returned as a CalViewProductDescriptor object.
		//
		//	ARGUMENTS:
		//		In:
		//			setName: Name of the data product to be created; if
		//					 empty (i.e. "") a default name will be created
		//

	unsigned subViewables() const;
		//
		//	PURPOSE:
		//		Return the number of subsidiary Viewables this Viewable
		//		contains
		//

	const std::vector<CalViewable *> &subViewable() const;
		//
		//	PURPOSE:
		//		Return a vector with pointers to subsidiary Viewables
		//		associated with this Viewables - can be used in conjunction
		//		with subViewables() to iterate over all available
		//		subsidiary Viewables and build of cascaded submenus in
		//		a GUI application

	virtual ~CalViewable();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor

protected:
	std::vector<CalViewable *>	_subViewable;
	std::string					_name, _whatIs;
}; // CalViewable()


//
//	CalViewServer:
//		Objects of class CalViewServer have a state which defines
//		the behavior of CalViewables and hold a vector with all
//		available CalViewable objects.
//
class CalViewServer {
public:
	CalViewServer();
		//
		//	PURPOSE:
		//		Create a CalViewServer object
		//

	CalViewState *state();
		//
		//	PURPOSE:
		//		Return a pointer to the state object of this CalViewServer
		//

	unsigned viewables() const;
		//
		//	PURPOSE:
		//		Return the number of CalViewables this CalViewServer knows
		//		about

	std::vector<const CalViewable *> viewable() const;
		//
		//	PURPOSE:
		//		Return a vector with pointers to Viewables known
		//		to this CalViewServer - can be used in conjunction
		//		with viewables() to iterate over all available
		//		Viewables and build a menu in a GUI application.
		//		The vector comprises only viewables which are applicable
		//		to the currently selected instrument (e.g. no redistribution
		//		function for OM).
		//

	~CalViewServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor

private:
	std::vector<CalViewable *>	_viewable;
	CalViewState			_state;
}; // CalViewServer()


#define calViewServer	calViewServerFunc()

CalViewServer &calViewServerFunc();	// single global instance of CalViewServer
									// object

#endif
