//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalF90cxxGlue.h	--- the Fortran90/C++ interface of the CAL
//
//	DESCRIPTION:
//		This file acts as the interface between the CAL API written in
//		Fortran90 and its implementation in C++. All platform-specifics
//		of the C++/F90 interface are centralized in this file and
//		CalF90cxxGlue.cc.
//
//	CHANGE HISTORY:
//		UL, 1997-07-25: 1.0.0: - first version
//		UL, 1998-??-??: 1.X.X: - many improvements
//		UL, 1999-04-01: 2.0.0: - new methods to set state from a reference
//								 to a block or a data set name
//		UL, 1999-06-08: 2.1.0: - added CAL_random()
//		UL, 1999-07-09: 2.2.0: - added state variables `ocb' and
//								 `temperatureCamera'
//							   - added CAL_rgsOffsetCorrect()
//		UL, 1999-08-13: 2.3.0: - new coordinate conversion routines added
//                             - added CAL_piTokeV
//		UL, 1999-08-19: 2.4.0: - dropped obsolete state variables `status'
//								 and `verbosityLevel'
//		UL, 1999-09-07: 2.4.1: - added rgsGetEventThresholds
//							   - added hasMiscellaneousData
//							   - changed I/F of rgsCtiCorrect (new arguments)
//							   - added CAL_rgsGetRedistribution
//							   - changed piToKev->piToeV
//		UL, 1999-09-17: 2.4.2: - changed arguments of CAL_rgsGetCrossPSF
//		UL, 1999-11-08: 2.4.3: - added CAL_getCamCoord2ToSacCoord
//		UL, 1999-11-23: 2.5.0: - added CAL_correctTime I/Fs
//							   - modified CAL_getCrossPsf I/F
//		UL, 2000-01-13: 2.5.1: - modified CAL_omDistortion/CAL_omGetColorTrans
//		UL, 2000-02-23: 2.6.0: - made CAL_getFilterTransmission
//									  CAL_getQuantumEfficiency
//								 functions
//		UL, 2000-03-09: 2.7.0: - added CAL_getFOVmap
//		UL, 2000-03-24: 2.8.0: - dropped time arguments of
//								 CAL_pn{Gain,Cti}Correct
//		UL, 2000-04-13: 3.3.0: - reconciled effective area I/F with C++ I/F
//		UL, 2000-04-25: 3.3.1: - additional quantities returned by
//									CAL_getModeParameters
//		UL, 2000-05-05: 3.2.0: - made CAL_getVignettingFactor a function
//		UL, 2000-05-11: 3.3.0: - added CAL_mosCtiCorrect()
//		UL, 2000-05-30: 3.4.0: - added alternative form of
//								 CAL_getRedistribution()
//							   - added CAL_getBinnedPIebounds()
//		UL, 2000-07-27: 3.5.0: - added CAL_omLEDtemplateMap()
//		UL, 2000-07-28: 3.6.0: - added CAL_getMOSoffsets()
//									   CAL_getMOSdarkFrameMap()
//		UL, 2000-09-20: 3.7.0: - added various PSF-related routines
//		UL, 2000-11-01: 3.8.0: - added various routines related to
//								 OM color transformation
//		UL, 2001-02-12: 3.9.0: - added various routines related to
//								 OM color transformation (in line with
//								 CAL HB 2.0)
//							   - added parameter empiricalLinearityCorrection
//								 to CAL_omPhotoNatural()
//							   - added CAL_omGetApertureRadius()
//		UL, 2001-03-09: 3.10.0:- added various new Boresight related stuff
//		UL, 2001-04-23: 3.11.0:- added new state variable 'auxiliaryParameter'
//		UL, 2001-06-01: 3.12.0:- added new routines:
//									o CAL_omGetPSF()
//									o CAL_omPsfEncircledEnergy()
//									o CAL_omPsfCircleRadius()
//		UL, 2001-12-05: 3.13.0:- added new routines:
//									o CAL_getBadPixelTableCode()
//									o CAL_getAduconvCode()
//		UL, 2002-03-05: 3.15.0:- new CAL_toDirectionCosineMatrix
//								 (intended to replace CAL_getBoresightMatrix)
//		UL, 2002-06-10: 3.16.0:- new CAL_getMiscellaneousDataCcd
//		UL, 2002-07-25: 3.17.0:- new CAL_pnAdditionalGainCtiCorrect
//									 CAL_pnGainTimingCorrect
//									 CAL_pnGainBurstCorrect
//		UL, 2002-08-16: 3.18.0:- various functions related to retrieval of
//								 pattern fraction data from CCF (e.g.
//								 CAL_getPatternFractionInEnergy)
//		RDS, 2003-10-03: 3.19.0:- added CAL_getLargePatterns to
//                   retrieve large pattern values from MOS QE files
//		UL/RDS, 2003-10-21: 3.20.0: - added CAL_omGrismFlux/Wavelength()
//		RDS, 2003-11-04: 3.21.0: - modified CAL_omInverseDistortion
//    RDS, 2003-11-27: 3.22.0: - added CAL_pnMedianMap, CAL_pnOffsetCorrValues,
//                               CAL_pnNoiseMap
//      RDS, 2007-01-25: 3.23.0: - added CAL_pnMasterOffset
//      RDS, 2007-02-16: 3.24.0: - added CAL_eVToPi
//      RDS, 2009-04-17: 3.25.0: - added CAL_setPosAngle
//      RDS, 2011-05-13: 3.26.0: - added CAL_pnSpatialCtiCorrect
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//		[3]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/f90/CalF90cxxGlue.h,v 1.78 2013/04/03 15:34:37 rsaxton Exp $
//

#ifndef CAL_F90CPP_H
#define CAL_F90CPP_H

#include "common.h"
#include "f90cxx.h"
#include "caloaldefs.h"
#include "Handle.h"
#include "Dal.h"
#include "PsfDataServer.h"
#include "CcdQuantumDataServer.h"
#include "OmPsfDataServer.h"


extern "C" void CXXFUNC(cal_openccf, CAL_openCCF)
							(const char *cifName, Integer3 length);


extern "C" void CXXFUNC(cal_setstate1, CAL_setState1)
							(const Integer3 *instrument,
							 const Integer3 *ccdChipId,
							 const Integer3 *ccdOutNode,
							 const Integer3 *modeId,
							 const Integer3 *filterId,
							 const Integer3 *accuracyLevel,
							 const Double *temperature,
							 const Double *temperatureCamera,
							 const Double *exposureStartTime,
							 const Double *exposureDuration,
							 const Integer3 *ocb,
							 const Logical *randomize,
							 const Logical *temperatureTracking,
							 const Double *auxiliaryParameter);

extern "C" void CXXFUNC(cal_setstatefromsciencefile,
						CAL_setStateFromScienceFile)
							(char *dataSetName, Integer3 length);

extern "C" void CXXFUNC(cal_setstate2, CAL_setState2)
							(char *dataSetName, Integer3 length);

extern "C" void CXXFUNC(cal_setstate3, CAL_setState3)
							(Handle<DataSet> *setP);

extern "C" void CXXFUNC(cal_setstate4, CAL_setState4)
							(Handle<Block> *blockP);

extern "C" void CXXFUNC(cal_getstate, CAL_setState)
							(Integer3 *instrument,
							 Integer3 *ccdChipId,
							 Integer3 *ccdOutNode,
							 Integer3 *modeId,
							 Integer3 *filterId,
							 Integer3 *accuracyLevel,
							 Double *temperature,
							 Double *temperatureCamera,
							 Double *exposureStartTime,
							 Double *exposureDuration,
							 Integer3 *ocb,
							 Logical *randomize,
							 Logical *temperatureTracking,
							 Double *auxiliaryParameter);


extern "C" void CXXFUNC(cal_setrandom, CAL_setRandom)
							(const Double *min,
							 const Double *max);


extern "C" void CXXFUNC(cal_setcalibrationfile, CAL_setCalibrationFile)
							(const char *calibrationFileName, Integer3 length);

/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct CalInfo {	// must match definitions of CalInfo type
							// in cal.f90
	char	instrument[10],	// instrument name
			calType[20],	// calibration file type
			fileName[256],	// full absolute name
			date[80],		// creation DATE
			valDate[80],	// VALDATE
			extSequ[80],	// EXTSEQU
			extSequId[80];	// EXTSEQID
	int		issue;			// file issue number
};

extern "C" void CXXFUNC(cal_ccfinfo, CAL_ccfInfo)
							(const char *calType,
							 struct CalInfo *, Integer3 calTypeLength);

extern "C" void CXXFUNC(cal_integertoreal, CAL_integerToReal)
							(const Dope1 *intVec,
							 Dope1 *realVec);

extern "C" void CXXFUNC(cal_realtointeger, CAL_realToInteger)
							(const Dope1 *realVec,
							 Dope1 *intVec);

#define RELEASEPROTO(fname, cname, N)		\
	extern "C" void CXXFUNC(fname, cname)	\
		(Dope##N *a1, Dope##N *a2, Dope##N *a3)

RELEASEPROTO(cal_releasememoryint3onedim, CAL_releaseMemoryInt3OneDim, 1);
RELEASEPROTO(cal_releasememorylogicalonedim, CAL_releaseMemoryLogicalOneDim,1);
RELEASEPROTO(cal_releasememoryreal2onedim, CAL_releaseMemoryReal2OneDim, 1);
RELEASEPROTO(cal_releasememoryint3twodim, CAL_releaseMemoryInt3TwoDim, 2);
RELEASEPROTO(cal_releasememoryreal2twodim, CAL_releaseMemoryReal2TwoDim, 2);

RELEASEPROTO(cal_releasememoryint1onedim, CAL_releaseMemoryInt1OneDim, 1);
RELEASEPROTO(cal_releasememoryint2onedim, CAL_releaseMemoryInt2OneDim, 1);
RELEASEPROTO(cal_releasememoryreal1onedim, CAL_releaseMemoryReal1OneDim, 1);
RELEASEPROTO(cal_releasememoryint2twodim, CAL_releaseMemoryInt2TwoDim, 2);
RELEASEPROTO(cal_releasememoryreal1twodim, CAL_releaseMemoryReal1TwoDim, 2);

RELEASEPROTO(cal_releasememoryint1threedim, CAL_releaseMemoryInt1ThreeDim, 3);

#undef RELEASEPROTO

extern "C" void CXXFUNC(cal_getbadpixellist, CAL_getBadPixelList)
							(Dope1 *rawX,
							 Dope1 *rawY,
							 Dope1 *type,
							 Dope1 *yextent,
							 Dope1 *uplinked,
							 const Integer2 *rawX0,
							 const Integer2 *rawY0,
							 const Integer2 *rawXsize,
							 const Integer2 *rawYsize);


extern "C" Logical CXXFUNC(cal_ispixelbad, CAL_isPixelBad)
							(const Integer2 *rawX,
							 const Integer2 *rawY,
							 Integer2 *yextent,
							 Integer2 *defect,
							 Logical *uplinked);


extern "C" void CXXFUNC(cal_getbadpixelmap, CAL_getBadPixelMap)
							(Dope2 *map,
							 const Integer2 *rawX0,
							 const Integer2 *rawY0,
							 const Integer2 *rawXsize,
							 const Integer2 *rawYsize);

//
//	---( CAL_getBoresight / CAL_withSacCoordOrientation /
//		 CAL_projectOntoTelCoord )------------------------------------------
//
extern "C" Handle<Boresight> CXXFUNC(cal_getboresight,
									 CAL_getBoresight)
							 (const Double *time);

extern "C" void CXXFUNC(cal_withsaccoordorientation,
						CAL_withSacCoordOrientation)
							(const Handle<Boresight> *boresight,
							 const SpacecraftAttitudeType *att);

extern "C" void CXXFUNC(cal_projectontotelcoord,
						CAL_projectOntoTelCoord)
							(const Handle<Boresight> *boresight,
							 const Dope1 *srcDirs,
							 Dope1 *theta,
							 Dope1 *phi);

extern "C" void CXXFUNC(cal_getboresightmatrix, CAL_getBoresightMatrix)
							(const Double *time, Dope2 *matrix);

extern "C" Dope2 CXXFUNC(cal_todirectioncosinematrix1,
						 CAL_toDirectionCosineMatrix1)
							(const Handle<Boresight> *boresight);

struct BSangles { double phi, theta, psi; };
extern "C" BSangles CXXFUNC(cal_getboresightangles, CAL_getBoresightAngles)
							(const Double *time);

extern "C" BSangles CXXFUNC(cal_toeulerangles, CAL_toEulerAngles)
							(const Handle<Boresight> *boresight);

extern "C" Dope2 CXXFUNC(cal_todirectioncosinematrix2,
						 CAL_toDirectionCosineMatrix2)
							(const BSangles *eulerAngles);

extern "C" void CXXFUNC(cal_euler321tomatrix, CAL_euler321toMatrix)
							(const BSangles *eulerAngles,
							 Double *matrix);

extern "C" void CXXFUNC(cal_getcamcoord2tosaccoord, CAL_getCamCoord2ToSacCoord)
							(const Double *time, Dope2 *matrix);

extern "C" Logical CXXFUNC(cal_hasmiscellaneousdata, CAL_hasMiscellaneousData)
							(const char *keyword,
							 Integer3 length);

extern "C" Logical CXXFUNC(cal_hasmiscellaneousdataccd,
						   CAL_hasMiscellaneousDataCcd)
							(const char *keywordStem,
							 Integer3 length);

extern "C" void CXXFUNC(cal_getmiscellaneousdata, CAL_getMiscellaneousData)
							(const char *keyword, Double *value,
							 Integer3 length);

extern "C" Double CXXFUNC(cal_getmiscellaneousdatavalue,
						  CAL_getMiscellaneousDataValue)
							(const char *keyword, Integer3 length);

extern "C" Double CXXFUNC(cal_getmiscellaneousdataccd,
						  CAL_getMiscellaneousDataCcd)
							(const char *keywordStem, Integer3 length);

/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct ContaminationData {	// must match definitions of
									// ContaminationDataType in CAL.f90
	int		TBS;			// to be specified
};

extern "C" void CXXFUNC(cal_getcontaminationdata, CAL_getContaminationData)
							(const double *time,
							 ContaminationData *contaminationData);


extern "C" void CXXFUNC(cal_getebounds, CAL_getEbounds)
							(Dope1 *channel,
							 Dope1 *Emin,
							 Dope1 *Emax);

extern "C" void CXXFUNC(cal_getebins, CAL_getEbins)
							(Dope1 *channel,
							 Dope1 *Emin,
							 Dope1 *Emax);


extern "C" void CXXFUNC(cal_getbinnedpiebounds, CAL_getBinnedPIebounds)
							(Dope1 *channel,
							 Dope1 *Emin,
							 Dope1 *Emax);


extern "C" void CXXFUNC(cal_getbadpixeltablecode, CAL_getBadPixelTableCode)
							(char *codeStr, Integer3 length);


extern "C" void CXXFUNC(cal_getaduconvcode, CAL_getAduconvCode)
							(char *codeStr, Integer3 length);


extern "C" void CXXFUNC(cal_getbackgroundmap, CAL_getBackgroundMap)
							(const Double *time,
							 Dope2 *map);


extern "C" void CXXFUNC(cal_getdarkframe, CAL_getDarkFrame)
							(Dope2 *map);


extern "C" void CXXFUNC(cal_getdiffusexbackground, CAL_getDiffuseXBackground)
							(const Double *time,
							 const Double *theta,
							 const Double *phi,
							 Dope1 *spectrum,
							 Dope1 *piVec);


extern "C" Double CXXFUNC(cal_getfiltertransmission1,
						  CAL_getFilterTransmission1)
							(const Double *E,
							 const Double *theta,
							 const Double *phi);

extern "C" Double CXXFUNC(cal_getfiltertransmission2,
						  CAL_getFilterTransmission2)
							(const Double *Efrom,
							 const Double *Eto,
							 const Double *theta,
							 const Double *phi);

extern "C" void CXXFUNC(cal_getfovmap, CAL_getFOVmap)
							(Dope2 *map,
							 const Integer2 *rawX0,
							 const Integer2 *rawY0,
							 const Integer2 *rawXsize,
							 const Integer2 *rawYsize);

/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct ModeParameterData {
	float	integrationTime,
			shiftTime;
	double  frameTime;
	float	cycleTime,
			timeResolution,
			dutyCycle,
			smearedFraction,
			pointSourcePileUp,
			limitingFlux;
	short	underscanX,
			overscanX,
			underscanY,
			overscanY,
			windowX0,
			windowY0,
			windowDX,
			windowDY;
};

extern "C" void CXXFUNC(cal_getmodeparameters, CAL_getModeParameters)
							(ModeParameterData *modeParameterData);


extern "C" void CXXFUNC(cal_getmosoffsets, CAL_getMOSoffsets)
							(Dope1 *offX, Dope1 *offY);


extern "C" void CXXFUNC(cal_getmosdarkframemap, CAL_getMOSdarkFrameMap)
							(Dope2 *map);


extern "C" void CXXFUNC(cal_getparticlebackground, CAL_getParticleBackground)
							(const Double *time,
							 const Double *theta,
							 const Double *phi,
							 Dope1 *spectrum,
							 Dope1 *piVec);


extern "C" void CXXFUNC(cal_getredistribution1, CAL_getRedistribution1)
							(const Double *eIn,
							 Dope1 *response,
							 Integer3 *pixeltype,
							 Integer2 *peakChannel);


extern "C" void CXXFUNC(cal_getredistribution2, CAL_getRedistribution2)
							(const Double *eIn,
							 const Dope1 *eAxisLoBounds,
							 const Dope1 *eAxisHiBounds,
							 Dope1 *response,
							 Integer3 *pixeltype,
							 Integer2 *peakChannel);


extern "C" Double CXXFUNC(cal_getvignettingfactor, CAL_getVignettingFactor)
							(const Double *E,
							 const Double *theta,
							 const Double *phi);


extern "C" Double CXXFUNC(cal_getquantumefficiency1, CAL_getQuantumEfficiency1)
							(const Double *E,
							 const Integer2 *rawX,
							 const Integer2 *rawY,
							 const Integer3 *patternId);

extern "C" Double CXXFUNC(cal_getquantumefficiency2, CAL_getQuantumEfficiency2)
							(const Double *Efrom,
							 const Double *Eto,
							 const Integer2 *rawX,
							 const Integer2 *rawY,
							 const Integer3 *patternId);

extern "C" Double CXXFUNC(cal_getlargepatternsize1, CAL_getLargePatternSize1)
							(const Double *E);

extern "C" Double CXXFUNC(cal_getlargepatternsize2, CAL_getLargePatternSize2)
                                                        (const Double *Efrom,
                                                         const Double *Eto);

extern "C" Double CXXFUNC(cal_geteffectivearea1, CAL_getEffectiveArea1)
							(const Double *E,
							 const Double *theta,
							 const Double *phi);

extern "C" Double CXXFUNC(cal_geteffectivearea2, CAL_getEffectiveArea2)
							(const Double *Efrom,
							 const Double *Eto,
							 const Double *theta,
							 const Double *phi);

extern "C" Double CXXFUNC(cal_rgsgeteffectivearea1, CAL_rgsGetEffectiveArea1)
							(const Double *E,
							 const Integer3 *order,
							 const Double *theta,
							 const Double *phi);

extern "C" Double CXXFUNC(cal_rgsgeteffectivearea2, CAL_rgsGetEffectiveArea2)
							(const Double *Efrom,
							 const Double *Eto,
							 const Integer3 *order,
							 const Double *theta,
							 const Double *phi);

extern "C" void CXXFUNC(cal_geteventpatterns, CAL_getEventPatterns)
							(Dope3 *patterns, Integer2 *eduThreshold);

extern "C" Double CXXFUNC(cal_curvevalueat, CAL_curveValueAt)
							(const Handle<CcdQuantumEfficiencyCurve> *curve,
							 const Double *abscissaValue);

extern "C" Double CXXFUNC(cal_curveweightedintegralover,
						  CAL_curveWeightedIntegralOver)
							(const Handle<CcdQuantumEfficiencyCurve> *curve,
							 const Double *fromValue, const Double *toValue);

extern "C" Handle<PatternFractionDataServer>
			  CXXFUNC(cal_getpatternfractionserver,
					  CAL_getPatternFractionServer)();

extern "C" void CXXFUNC(cal_patternfractionat, CAL_patternFractionAt)
							(const Handle<PatternFractionDataServer> *pfds,
							 const Integer2 *chipX, const Integer2 *chipY);

extern "C" Handle<CcdQuantumEfficiencyCurve>
			  CXXFUNC(cal_getpatternfractiontotalqe,
					  CAL_getPatternFractionTotalQE)
							(const Handle<PatternFractionDataServer> *pfds);

extern "C" Handle<CcdQuantumEfficiencyCurve>
			  CXXFUNC(cal_getpatternfractioninenergy,
					  CAL_getPatternFractionInEnergy)
							(const Handle<PatternFractionDataServer> *pfds,
							 const Integer3 *grade);

extern "C" Dope1
			  CXXFUNC(cal_getpatternfractioninchannel,
					  CAL_getPatternFractionInChannel)
							(const Handle<PatternFractionDataServer> *pfds,
							 const Integer3 *grade);

extern "C" void CXXFUNC(cal_gethkwindows, CAL_getHKwindows)
							(Dope1 *windowList);

extern "C" void CXXFUNC(cal_correcttime1, CAL_correctTime1)
							(const Dope1 *timeCoordinate);
extern "C" void CXXFUNC(cal_correcttime2, CAL_correctTime2)
							(const Dope1 *timeCoordinate,
							 const Dope1 *rawY);
extern "C" void CXXFUNC(cal_correcttime3, CAL_correctTime3)
							(const Dope1 *timeCoordinate,
							 const Dope1 *rawY,
							 Integer2 *sourceY);
//
// Time jumps
//
extern "C" Integer3 CXXFUNC(cal_timejumptol, CAL_timeJumpTol) 
							(const Integer3 *rev);

//
//	PSF
//
extern "C" Handle<Psf> CXXFUNC(cal_getpsf, CAL_getPSF)
							(const Double *E,
							 const Double *theta,
							 const Double *phi);

extern "C" Double CXXFUNC(cal_psfpointvalue, CAL_psfPointValue)
							(const Handle<Psf> *psf,
							 const Double *x,
							 const Double *y);

extern "C" void CXXFUNC(cal_psfsetposangle, CAL_psfSetPosAngle)
							(const Handle<Psf> *psf,
							 const Double *posangle);

extern "C" Logical CXXFUNC(cal_psfbinsizechangeable, CAL_psfBinSizeChangeable)
							(const Handle<Psf> *psf);

/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct PsfBinSize {
	Double	binX, binY;
};
extern "C" void CXXFUNC(cal_psfsetbinsize, CAL_psfSetBinSize)
							(const Handle<Psf> *psf,
							 const PsfBinSize *binSize);

extern "C" PsfBinSize CXXFUNC(cal_psfbinsize, CAL_psfBinSize)
							(const Handle<Psf> *psf);

/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct PsfValidityRanges {
	Double	eMin, eMax, thetaMin, thetaMax, phiMin, phiMax;
};
extern "C" PsfValidityRanges CXXFUNC(cal_psfvalidityranges,
									 CAL_psfValidityRanges)
							();
/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct PsfWindow {
	Double	halfX, halfY;
};
extern "C" void CXXFUNC(cal_psfimage, CAL_psfImage)
							(const Handle<Psf> *psf,
							 const PsfWindow *window,
							 Dope2 *image);

extern "C" PsfWindow CXXFUNC(cal_psfenboxedenergy, CAL_psfEnboxedEnergy)
							(const Handle<Psf> *psf,
							 const Double *fraction,
							 const Double *aspectRatio);

extern "C" Double CXXFUNC(cal_psfencircledenergy, CAL_psfEncircledEnergy)
							(const Handle<Psf> *psf,
							 const Double *fraction);

extern "C" Double CXXFUNC(cal_psfcentreoffset, CAL_psfCentreOffset)
                                                        (const Handle<Psf> *psf)
;

//
//	coordinate conversion routines
//

//	deprecated
//		PIXCOORD1 -> CAMCOORD1
//		PIXCOORD1 -> CAMCOORD2
//		PIXCOORD1 -> SACCORD
extern "C" void CXXFUNC(cal_rawxy2mm, CAL_rawXY2mm)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 Dope1 *Xmm,
							 Dope1 *Ymm,
							 Dope1 *Zmm,
							 const Integer3 *coordType);

//	PIXCOORD0 -> PIXCOORD1
extern "C" void CXXFUNC(cal_pixcoord0topixcoord1, CAL_pixCoord0ToPixCoord1)
							(const Dope1 *telX,
							 const Dope1 *telY,
							 Dope1 *rawX,
							 Dope1 *rawY);


//	PIXCOORD1 <-> CHIPCOORD
extern "C" void CXXFUNC(cal_pixcoord1tochipcoord, CAL_pixCoord1ToChipCoord)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 Dope1 *chipX,
							 Dope1 *chipY,
							 const Dope1 *nodes);

extern "C" void CXXFUNC(cal_chipcoordtopixcoord1, CAL_chipCoordToPixCoord1)
							(const Dope1 *chipX,
							 const Dope1 *chipY,
							 Dope1 *rawX,
							 Dope1 *rawY,
                             Dope1 *nodes);

//	CHIPCOORD <-> CAMCOORD1
extern "C" void CXXFUNC(cal_chipcoordtocamcoord1, CAL_chipCoordToCamCoord1)
							(const Dope1 *chipX,
							 const Dope1 *chipY,
							 Dope1 *xmm1,
							 Dope1 *ymm1,
							 Dope1 *zmm1);

extern "C" void CXXFUNC(cal_camcoord1tochipcoord, CAL_camCoord1ToChipCoord)
							(const Dope1 *xmm1,
							 const Dope1 *ymm1,
							 const Dope1 *zmm1,
							 Dope1 *chipX,
							 Dope1 *chipY,
							 Dope1 *ccds);

//	CAMCOORD1 <-> CAMCOORD2
extern "C" void CXXFUNC(cal_camcoord1tocamcoord2, CAL_camCoord1ToCamCoord2)
							(const Dope1 *xmm1,
							 const Dope1 *ymm1,
							 const Dope1 *zmm1,
							 Dope1 *xmm2,
							 Dope1 *ymm2,
							 Dope1 *zmm2);

extern "C" void CXXFUNC(cal_camcoord2tocamcoord1, CAL_camCoord2ToCamCoord1)
							(const Dope1 *xmm2,
							 const Dope1 *ymm2,
							 const Dope1 *zmm2,
							 Dope1 *xmm1,
							 Dope1 *ymm1,
							 Dope1 *zmm1);

//	CAMCOORD2 <-> SACOORD
extern "C" void CXXFUNC(cal_camcoord2tosaccoord, CAL_camCoord2ToSacCoord)
							(const Dope1 *xmm2,
							 const Dope1 *ymm2,
							 const Dope1 *zmm2,
							 Dope1 *xmmSc,
							 Dope1 *ymmSc,
							 Dope1 *zmmSc);

extern "C" void CXXFUNC(cal_saccoordtocamcoord2, CAL_sacCoordToCamCoord2)
							(const Dope1 *xmmSc,
							 const Dope1 *ymmSc,
							 const Dope1 *zmmSc,
							 Dope1 *xmm2,
							 Dope1 *ymm2,
							 Dope1 *zmm2);


//	SACCOORD <-> ROWCOORD
extern "C" void CXXFUNC(cal_saccoordtorowcoord, CAL_sacCoordToRowCoord)
							(const Dope1 *xmmSc,
							 const Dope1 *ymmSc,
							 const Dope1 *zmmSc,
							 Dope1 *beta,
							 Dope1 *chi);

extern "C" void CXXFUNC(cal_rowcoordtosaccoord, CAL_rowCoordToSacCoord)
							(const Dope1 *beta,
							 const Dope1 *chi,
							 Dope1 *xmmSc,
							 Dope1 *ymmSc,
							 Dope1 *zmmSc);


//	CAMCOORD2 <-> TELCOORD
extern "C" void CXXFUNC(cal_camcoord2totelcoord, CAL_camCoord2ToTelCoord)
							(const Dope1 *xmm2,
							 const Dope1 *ymm2,
							 const Dope1 *zmm2,
							 Dope1 *theta,
							 Dope1 *phi);

extern "C" void CXXFUNC(cal_telcoordtocamcoord2, CAL_telCoordToCamCoord2)
							(const Dope1 *theta,
							 const Dope1 *phi,
							 Dope1 *xmm2,
							 Dope1 *ymm2,
							 Dope1 *zmm2);


extern "C" void CXXFUNC(cal_offsetcorrect, CAL_offsetCorrect)
							(const Dope1 *time,
							 const Dope1 *eIn,
							 Dope1 *eOut,
							 const Dope1 *nodes);


extern "C" void CXXFUNC(cal_gaincorrect, CAL_gainCorrect)
							(const Dope1 *time,
							 const Dope1 *eIn,
							 Dope1 *eOut,
							 const Dope1 *nodes);

extern "C" void CXXFUNC(cal_pitoev, CAL_piToeV)
                            (const Dope1 *eIn,
                             Dope1 *eOut);

extern "C" void CXXFUNC(cal_evtopi, CAL_eVToPi)
                            (const Dope1 *eIn,
                             Dope1 *eOut);

//
//	MOS specific
//
extern "C" void CXXFUNC(cal_mosgaincorrect, CAL_mosGainCorrect)
							(const Dope1 *time,
							 const Dope1 *pattern,
							 const Dope1 *eIn,
							 Dope1 *eOut);

extern "C" void CXXFUNC(cal_moscticorrect, CAL_mosCtiCorrect)
							(const Dope1 *time,
							 const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *eIn,
							 Dope1 *eOut);


//
//	EPN specific
//
extern "C" void CXXFUNC(cal_pngaincorrect, CAL_pnGainCorrect)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *eIn,
							 Dope1 *eOut);

extern "C" void CXXFUNC(cal_pncticorrect, CAL_pnCtiCorrect)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *frame,
							 const Dope1 *eIn,
							 Dope1 *eOut);

extern "C" void CXXFUNC(cal_pnadditionalgaincticorrect,
						CAL_pnAdditionalGainCtiCorrect)
							(Dope1 *,
							 const Dope1 *,
							 const Double *,
							 Dope1 *,
							 const Logical *,
							 const Logical *,
							 const Logical *,
							 const Logical *,
							 const Logical *);

extern "C" void CXXFUNC(cal_pnspatialcticorrect,
                                                CAL_pnSpatialCtiCorrect)
                                                        (Dope1 *,
                                                         const Dope1 *,
                                                         const Dope1 *);

extern "C" void CXXFUNC(cal_pngaintimingcorrect, CAL_pnGainTimingCorrect)
							(Dope1 *);

extern "C" void CXXFUNC(cal_pngainburstcorrect, CAL_pnGainBurstCorrect)
							(Dope1 *);

extern "C" void CXXFUNC(cal_pngaineffcorrect, CAL_pnGainEffCorrect)
							(Dope1 *);

extern "C" void CXXFUNC(cal_pngainffcorrect, CAL_pnGainFfCorrect)
							(Dope1 *);

extern "C" Double CXXFUNC(cal_pnreemissionthreshold, CAL_pnReEmissionThreshold)
							(const Double *energy,
							 const Integer2 *rawY);

extern "C" Double CXXFUNC(cal_pnratedependentphaparam, 
                            cal_pnRateDependentPhaParam)();

extern "C" void CXXFUNC(cal_pnmedianmap, CAL_pnMedianMap)
							(Dope2 *medmap);

extern "C" void CXXFUNC(cal_pnmasteroffset, CAL_pnMasterOffset)
							(Dope2 *masteroffset);

extern "C" void CXXFUNC(cal_pnmasterclosedoffset, CAL_pnMasterClosedOffset)
						(Dope2 *masteroffclosed,
                                                 Dope1 *xrlCoeff);

extern "C" void CXXFUNC(cal_pnoffsetcorrvalues, CAL_pnOffsetCorrValues)
							(Dope1 *hlo,
               Dope1 *hhi,
               Dope1 *val);

extern "C" void CXXFUNC(cal_pnnoisemap, CAL_pnNoiseMap)
							(Dope1 *rawX,
               Dope3 *noisemap);

//
//	RGS specific
//
extern "C" void CXXFUNC(cal_rgsoffsetcorrect, CAL_rgsOffsetCorrect)
							(const Dope1 *time,
							 const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *shapes,
							 const Dope1 *nodes,
							 const Dope1 *eIn,
							 Dope1 *eOut);

extern "C" void CXXFUNC(cal_rgscticorrect, CAL_rgsCtiCorrect)
							(const Dope1 *time,
							 const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *nodes,
							 const Dope1 *shape,
							 const Dope1 *separation,
							 const Dope1 *eIn,
							 Dope1 *eOut);


extern "C" void CXXFUNC(cal_rgsgetcrosspsf, CAL_rgsGetCrossPSF)
	                       (const Double *time,
							const Double *beta,
							const Double *sourceXdisp,
							const Dope1 *xAngles,
							const Double *binWidth,
							Dope1 *distribution);


extern "C" void CXXFUNC(cal_rgsgeteventthresholds, CAL_rgsGetEventThresholds)
							(Dope1 *rejectionThresholds,
							 Dope1 *acceptanceThresholds,
							 const Integer1 *node);

extern "C" void CXXFUNC(cal_rgsgetlsf, CAL_rgsGetLSF)
							(const Double *lambda,
							 const Double *theta,
							 const Double *phi,
							 const Integer3 *diffOrder,
							 const Double *lambdaMin,
							 const Double *lambdaMax,
							 Dope1 *lambdaGrid,
							 Dope1 *lsfProf);


extern "C" void CXXFUNC(cal_rgsgetredistribution, CAL_rgsGetRedistribution)
							(const Double *E,
							 const Integer2 *rawX,
							 const Integer2 *rawY,
							 const Double *rejectionThreshold,
							 Dope1 *redistribution,
							 Dope1 *channel);


extern "C" void CXXFUNC(cal_rgsgetrefinedwavescale, CAL_rgsGetRefinedWaveScale)
							(const Double *time,
							 const Dope1 *order,
							 Dope1 *wavelength);


//
//	OM specific
//
extern "C" void CXXFUNC(cal_omdistortion, CAL_omDistortion)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 Dope1 *deltaX,
							 Dope1 *deltaY,
							 Dope1 *sigmaX,
							 Dope1 *sigmaY,
							 const Integer3 *calcMode,
							 const Integer3 *nInterPoints);

extern "C" void CXXFUNC(cal_ominversedistortion, CAL_omInverseDistortion)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 Dope1 *deltaX,
							 Dope1 *deltaY,
							 Dope1 *sigmaX,
							 Dope1 *sigmaY,
							 const Integer3 *calcMode,
							 const Integer3 *nInterPoints);

extern "C" void CXXFUNC(cal_omgetgrismflux, CAL_omGetGrismFlux)
						   	 (Dope1 *lamda,
							  Dope1 *isf,
						   	  Dope1 *error);

extern "C" void CXXFUNC(cal_omgetgrismwavelength, CAL_omGetGrismWavelength)
							 (Dope1 *lamda);

extern "C" Handle<OmColorTransformator> CXXFUNC(cal_omgetcolortransformator,
												CAL_omGetColorTransformator)
							(const Integer3 *filter1,
							 const Integer3 *filter2);

extern "C" Integer1 CXXFUNC(cal_omcolortransbranches,
							CAL_omColorTransBranches)
							(const Handle<OmColorTransformator> *colorTrans);

extern "C" void CXXFUNC(cal_omcolortranssetbranch,
						CAL_omColorTransSetBranch)
							(const Handle<OmColorTransformator> *colorTrans,
							 const Integer1 *branch);

extern "C" Integer1 CXXFUNC(cal_omcolortransvalidityranges,
							CAL_omColorTransValidityRanges)
							(const Handle<OmColorTransformator> *colorTrans);
/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct IntervalT {
	Double	lower, upper;
};
extern "C" IntervalT CXXFUNC(cal_omcolortransvalidityrange,
							 CAL_omColorTransValidityRange)
							(const Handle<OmColorTransformator> *colorTrans,
							 const Integer1 *n);


struct ValueErrorT {
	Double	value, error;
};
extern "C" ValueErrorT CXXFUNC(cal_omstandardcolor, CAL_omStandardColor)
							(const Handle<OmColorTransformator> *colorTrans,
							 const Double *magnitude1,
							 const Double *magnitude2);

extern "C" ValueErrorT CXXFUNC(cal_omstandardmagnitude,
							   CAL_omStandardMagnitude)
							(const Handle<OmColorTransformator> *colorTrans,
							 const ValueErrorT *stdColor,
							 const Double *magnitude);

extern "C" OmColorTransformator::OmFluxConvF CXXFUNC(cal_omgetfluxconvfactors,
                               CAL_omGetFluxConvFactors)
                               ();

extern "C" void CXXFUNC(cal_omcolortransform, CAL_omColorTransform)
							(const Integer3 *filter1,
							 const Integer3 *filter2,
							 const Double *magnitude1,
							 const Double *magnitude2,
							 Double *stdColor12,
							 Double *stdMagnitude2,
							 Double *stdColor12Error,
							 Double *stdMagnitude2Error);


/* AI gcc4 porting
 *  ‘typedef’ was ignored in this declaration => remove typedef
 */
struct OMframeParameters {	// must match definitions of corresponding
									// type in cal.f90
		Double nHorizontalPixels,
               nVerticalPixels,
			   verticalTransferTime,
			   horizontalTransferTime,
			   nVerticalPixelsImage,
			   verticalHorizontalSwitchTime,
			   horizontalVerticalSwitchTime,
			   redundantCcdOffset;
};

extern "C" Double CXXFUNC(cal_omgetapertureradius, CAL_omGetApertureRadius)();

extern "C" void CXXFUNC(cal_omgetframeparameters, CAL_omGetFrameParameters)
							(struct OMframeParameters *frameParameters);

extern "C" void CXXFUNC(cal_omgetdegradationcoeffs, CAL_omGetDegradationCoeffs)
                            (Dope1 *degradCoeff);


extern "C" Handle<OmPsf> CXXFUNC(cal_omgetpsf, CAL_omGetPSF)
							(const Double *cenX,
							 const Double *cenY,
							 const Double *countFrameRateRatio);

extern "C" Double CXXFUNC(cal_ompsfencircledenergy, CAL_omPsfEncircledEnergy)
							(const Handle<OmPsf> *psf,
							 const Double *radius);

extern "C" Double CXXFUNC(cal_ompsfcircleradius, CAL_omPsfCircleRadius)
							(const Handle<OmPsf> *psf,
							 const Double *fraction);

extern "C" void CXXFUNC(cal_omgetpsfmap, CAL_omGetPSFmap)
							(const Double *rawX,
							 const Double *rawY,
							 const Integer3 *nPixelsNegX,
							 const Integer3 *nPixelsPosX,
							 const Integer3 *nPixelsNegY,
							 const Integer3 *nPixelsPosY,
							 const Integer3 *xRes,
							 const Integer3 *yRes,
							 const Double *countFrameRateRatio,
							 Dope2 *PSFmap);


extern "C" void CXXFUNC(cal_omgetplatescale, CAL_omGetPlateScale)
							(const Double *lambda,
							 Double *plateScale,
							 Double *errorX,
							 Double *errorY);


extern "C" void CXXFUNC(cal_omlargesensevariation, CAL_omLargeSenseVariation)
							(const Integer3 *x0,
							 const Integer3 *y0,
							 const Integer3 *xSize,
							 const Integer3 *ySize,
							 const Double *lambda,
							 Dope2 *senseMap,
							 Double *averageFlatField,
							 Double *averageFlatFieldError,
							 Double *pixelError);


extern "C" void CXXFUNC(cal_omledtemplatemap, CAL_omLEDtemplateMap)
							(const Integer3 *x0,
							 const Integer3 *y0,
							 const Integer3 *xSize,
							 const Integer3 *ySize,
							 Dope2 *senseMap,
							 Double *average,
							 Double *averageError,
							 Double *pixelError);


extern "C" void CXXFUNC(cal_omphotomagnitude, CAL_omPhotoMagnitude)
							(const Double *naturalRate,
							 Double *magnitude,
							 const Double *naturalRateErrorMinus,
							 const Double *naturalRateErrorPlus,
							 Double *magnitudeErrorMinus,
							 Double *magnitudeErrorPlus);


extern "C" Double CXXFUNC(cal_omnaturalmagnitude, CAL_omNaturalMagnitude)
							(const Double *naturalRate,
							 const Double *naturalRateErrorMinus,
							 const Double *naturalRateErrorPlus,
							 Double *magnitudeErrorMinus,
							 Double *magnitudeErrorPlus);


extern "C" Double CXXFUNC(cal_omphotonatural0, CAL_omPhotoNatural0)
							(const Double *rawRate,
							 const Double *frameTime,
							 const Double *deadFrac,
							 const Double *eventLoss,
							 const Logical *extendedSource,
							 const Double *rawRateErrorMinus,
							 const Double *rawRateErrorPlus,
							 Double *naturalRateErrorMinus,
							 Double *naturalRateErrorPlus);


struct ValueAsymmetricErrorT {
	Double	value, errorPlus, errorMinus;
};
extern "C" ValueAsymmetricErrorT
					CXXFUNC(cal_omphotonatural1, CAL_omPhotoNatural1)
							(const ValueAsymmetricErrorT *rawRate,
							 const Double *frameTime,
							 const Double *deadFrac,
							 const Double *eventLoss,
							 const Logical *extendedSource,
							 const Logical *empiricalLinearityCorrection);


extern "C" void CXXFUNC(cal_ompixelsensevariation, CAL_omPixelSenseVariation)
							(const Integer3 *x0,
							 const Integer3 *y0,
							 const Integer3 *xSize,
							 const Integer3 *ySize,
							 const Double *countRateFrameRatio,
							 Dope2 *senseMap,
							 Double *averageFlatField,
							 Double *averageFlatFieldError,
							 Double *pixelError);


//
//	translations from instrument/mode/filter names into integer Ids
//
extern "C" Integer3 CXXFUNC(instidtoint, instIdToInt)
							(const char *name, Integer3 length);
extern "C" Integer3 CXXFUNC(modeidtoint, modeIdToInt)
							(const char *name, Integer3 length);
extern "C" Integer3 CXXFUNC(datamodeidtoint, dataModeIdToInt)
							(const char *name, Integer3 length);
extern "C" Integer3 CXXFUNC(filteridtoint, filterIdToInt)
							(const char *name, Integer3 length);

extern "C" void CXXFUNC(instidtostring, instIdToString)
							(char *name,
						     Integer3 length,
							 const Integer3 *id);
extern "C" void CXXFUNC(modeidtostring, modeIdToString)
							(char *name,
						     Integer3 length,
							 const Integer3 *id);
extern "C" void CXXFUNC(datamodeidtostring, dataModeIdToString)
							(char *name,
						     Integer3 length,
							 const Integer3 *id);
extern "C" void CXXFUNC(filteridtostring, filterIdToString)
							(char *name,
						     Integer3 length,
							 const Integer3 *id);

extern "C" Double CXXFUNC(stringtotimetag, stringToTimeTag)
							(const char *name,
							 Integer3 length);

#endif
