//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					  (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalCommon.cc --- implementation of class methods in CalCommon.h
//
//	DESCRIPTION:
//		Implementation of class methods in CalCommon.h
//
//	CHANGE HISTORY:
//		UL, 1999-03-29: 1.0.0: - first version
//		UL, 1999-09-02: 1.1.0: - added BinnedAxis class family
//		UL, 1999-10-27: 1.2.0: - constructurs of NonUniformilyBinnedAxis
//								 take upper bound of last bin explicitely
//		UL, 2000-05-30: 1.3.0: - added new NonUniformilyBinnedAxis constructor
//		UL, 2001-05-04: 1.4.0: - bug fix in Spectrum::operator=()
//
//	AUTHOR:
//		U. Lammers, Mon Mar 29 15:09:59 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CalCommon.cc,v 1.24 2011/01/11 16:18:02 rsaxton Exp $
//

#include <algorithm>	//	rotate
#include "CalCommon.h"
#include "StringPool.h"

using namespace std;

//
//	Cal accuracy
//
static const char *__acclevels[] = { "LOW", "MEDIUM", "EXTENDED", "ELLBETA", "HIGH", 0 };
//static const char *__acclevels[] = { "LOW", "MEDIUM", "EXTENDED", "HIGH", 0 };
static const StringPool _acclevels(__acclevels, StringPool::ABORT);

const string &toString(AccuracyLevel accuracyLevel)
{
	return _acclevels[UNT(accuracyLevel)];
} // toString()

static const AccuracyLevel _accLevs[] = { LOW, MEDIUM, EXTENDED, ELLBETA, HIGH };
//static const AccuracyLevel _accLevs[] = { LOW, MEDIUM, EXTENDED, HIGH };
const vector<AccuracyLevel> accuracyChoices(_accLevs,
											_accLevs+NUM_OF(_accLevs));

namespace std {
ostream &operator<<(ostream &os, AccuracyLevel accuracyLevel)
{
	os<<INT(accuracyLevel)<<" ["<<toString(accuracyLevel)<<"]";
	return os;
} // operator<<
};

//
//	-------------( Spectrum )---------------------------------------------
//
Spectrum &Spectrum::operator=(double rhs)
{
	unsigned N = this->size();
	for (unsigned n=0; n<N; (*this)[n++] = rhs){};

	return *this;
} // Spectrum::operator=


Spectrum &Spectrum::operator=(const Spectrum &rhs)
{
	if (this != &rhs) {
		_momentSet = rhs._momentSet;
		_moment1 = rhs._moment1;
		vector<double>::operator=(rhs);
	}

	return *this;
} // Spectrum::operator=()


Spectrum &Spectrum::operator+=(const Spectrum &rhs)
{
	unsigned N = rhs.size();
	if (this->size() != N)
		errstr<<"Spectra have different sizes - they cannot be added"
			  <<Sas::error("DifferentSizedSpectra");
	vector<double> &v = *this;
	for (unsigned n=0; n<N; ++n)
		v[n] += rhs[n];

	return *this;
} // Spectrum::operator+=()


Spectrum &Spectrum::normalize(double normalization)
{
	double actnorm = norm();
	if (actnorm > 0.) {
		double norm = normalization/actnorm;
		vector<double> &v = *this;
		unsigned nChan = v.size();
		for (unsigned n=0; n<nChan; ++n) {
			if (v[n] < 0.) {
				errstr<<"Spectrum at channel #"<<n<<" is negative ("
					  <<v[n]<<") - setting it to 0"
					  <<Sas::message(Msg::Lib, Msg::Verbose);
				v[n] = 0.;
			}
			v[n] *= norm;
		}
	}

	return *this;
} // Spectrum::normalize()


double Spectrum::norm() const
{
	const vector<double> &v = *this;
	unsigned nChan = v.size();
	double sum = 0.;
	for (unsigned n=0; n<nChan; ++n)
		if (v[n] > 0.)
			sum += v[n];

	return sum;
} // Spectrum::normalize()


void Spectrum::rotate(int channels)
{
	::rotate(begin(),
		     channels<=0 ? begin()+ -channels%size()
					     : end()-(channels-1)%size()-1,
			 end());
} // Spectrum::rotate()


double Spectrum::firstMoment() const
{
	if (_momentSet)
		return _moment1;

	unsigned N = size();
	double t = 0;
	for (unsigned n=0; n<N; ++n)
		t += (DBL(n)+1.)*(*this)[n];

	return t;
} // Spectrum::firstMoment()


unsigned Spectrum::peakChannel() const
{
	unsigned peakChan = 0;
	const vector<double> &v = *this;
	unsigned nChan = v.size();
	double val = -1.;
	for (unsigned n=0; n<nChan; ++n)
		if (v[n] > val) {
			val = v[n];
			peakChan = n;
		}

	return peakChan;
} // Spectrum::peakChannel


//
//	-------------( BinnedAxis )-------------------------------------------
//
BinnedAxis::BinnedAxis()
{
} // BinnedAxis::BinnedAxis()


BinnedAxis::BinnedAxis(double lower, double upper)
{
	if (lower < upper) {
		_lower = lower;
		_upper = upper;
	} else
		errstr<<"Lower bound ("<<lower<<") is greater than upper bound ("
			  <<upper<<")"
			  <<Sas::error("LowerGreaterThanUpper");
} // BinnedAxis::BinnedAxis()


BinnedAxis::~BinnedAxis()
{
} // BinnedAxis::~BinnedAxis()


UniformilyBinnedAxis::UniformilyBinnedAxis(double lower,
										   double upper, unsigned bins)
	: BinnedAxis(lower, upper)
{
	if (!bins)
		errstr<<"The number of bins cannot be zero"<<Sas::error("ZeroBins");

	_bins = bins;
	_binWidth = (upper - lower)/DBL(bins);
	_v = new unsigned[_bins];
	for (unsigned n=0; n<_bins; _v[n++] = 0){};
} // UniformilyBinnedAxis::UniformilyBinnedAxis()


UniformilyBinnedAxis::UniformilyBinnedAxis(double lower, unsigned bins,
										   double binWidth)
	: BinnedAxis(lower, bins * binWidth + lower)
{
	if (!bins)
		errstr<<"The number of bins cannot be zero"<<Sas::error("ZeroBins");
	if (binWidth < 0.)
		errstr<<"The bin width cannot be negative ("<<binWidth<<")"
			  <<Sas::error("NegativeBinWidth");

	_binWidth = binWidth;
	_bins = bins;
	_v = new unsigned[_bins];
} // UniformilyBinnedAxis::UniformilyBinnedAxis()


UniformilyBinnedAxis::~UniformilyBinnedAxis()
{
	delete [] _v;
	_v = 0;
} // UniformilyBinnedAxis::~UniformilyBinnedAxis()


unsigned UniformilyBinnedAxis::bins() const
{
	return _bins;
} // UniformilyBinnedAxis::bins()


Bin UniformilyBinnedAxis::bin(unsigned index) const
{
	if (index >= _bins)
		throw NoSuchBin();

	return Bin(DBL(index) * _binWidth + _lower,
			   DBL(index + 1) * _binWidth + _lower);
} // UniformilyBinnedAxis::bin()


Bin UniformilyBinnedAxis::bin(double value) const
{
	return bin(index(value));
} // UniformilyBinnedAxis::bin()


unsigned &UniformilyBinnedAxis::operator[](unsigned index)
{
	if (index >= _bins)
		throw NoSuchBin();

	return _v[index];
} // UniformilyBinnedAxis::operator[]


unsigned UniformilyBinnedAxis::index(double value) const
{
	if (value < _lower || value >= _upper)
		throw NoSuchBin();

	return UNT((value - _lower)/_binWidth);
} // UniformilyBinnedAxis::index()


unsigned &UniformilyBinnedAxis::operator[](double value)
{
	return _v[index(value)];
} // UniformilyBinnedAxis::operator[]


//
// -------------------( NonUniformilyBinnedAxis )----------------------------
//
template<class rvtype, class rtype>
void NonUniformilyBinnedAxis::init(const rvtype &lowerBounds,
								   bool withUpperBound, rtype upperBound)
{
	unsigned minSize = withUpperBound ? 1 : 2;
	if (lowerBounds.size() < minSize)
		errstr<<"Size of vector for initializing NonUniformilyBinnedAxis "
				"must be greater or equal to "<<minSize
			  <<Sas::error("InvalidInput");

	double min = lowerBounds[0] - 1.;
	unsigned N = lowerBounds.size();
	_lower = lowerBounds[0];
	_upper = lowerBounds[N-1];

	_lowerBounds.reserve(N+1);
	for (unsigned n=0; n<N; ++n)
		if (lowerBounds[n] > min) {
			_lowerBounds.push_back(lowerBounds[n]);
			min = _lowerBounds[n];
		} else
			errstr<<"Lower bound of bin #"<<n+1<<" ("<<lowerBounds[n]<<") is "
					"less than or equal to lower bound of preceding bin ("
				  <<min<<")"<<Sas::error("InvalidInput");

	_bins = N-1; // last value of _lowerBounds is upper bound of last bin
	if (withUpperBound) {
		_bins++;
		if (min > upperBound)
			errstr<<"Lower bound of bin #"<<N+1<<" ("<<min<<") is "
					"greater or equal to its upper bound ("
				  <<upperBound<<")"<<Sas::error("InvalidInput");
		_lowerBounds.push_back(upperBound);
		_upper = upperBound;
	}

	_v = new unsigned[_bins];
	for (unsigned n=0; n<_bins; _v[n++] = 0){};
} // NonUniformilyBinnedAxis::init()

NonUniformilyBinnedAxis::NonUniformilyBinnedAxis() : BinnedAxis(),
	_v(0), _bins(0)
{
} // NonUniformilyBinnedAxis::NonUniformilyBinnedAxis()


NonUniformilyBinnedAxis::NonUniformilyBinnedAxis(
									const CalReal32Vector &lowerBounds,
									bool withUpperBound, real32 upperBound)
	: BinnedAxis(0., 1.)
{
	init(lowerBounds, withUpperBound, upperBound);
} // NonUniformilyBinnedAxis::NonUniformilyBinnedAxis()


NonUniformilyBinnedAxis::NonUniformilyBinnedAxis(
									const CalReal64Vector &lowerBounds,
									bool withUpperBound, real64 upperBound)
	: BinnedAxis(0., 1.)
{
	init(lowerBounds, withUpperBound, upperBound);
} // NonUniformilyBinnedAxis::NonUniformilyBinnedAxis()


unsigned NonUniformilyBinnedAxis::bins() const
{
	return _bins;
} // NonUniformilyBinnedAxis::bins()


Bin NonUniformilyBinnedAxis::bin(unsigned index) const
{
	if (index >= _bins)
		throw NoSuchBin();

	return Bin(_lowerBounds[index], _lowerBounds[index+1]);
} // NonUniformilyBinnedAxis::bin()


Bin NonUniformilyBinnedAxis::bin(double value) const
{
	return bin(index(value));
} // NonUniformilyBinnedAxis::bin()


unsigned &NonUniformilyBinnedAxis::operator[](unsigned index)
{
	if (index >= _bins)
		throw NoSuchBin();

	return _v[index];
} // NonUniformilyBinnedAxis::operator[]


unsigned NonUniformilyBinnedAxis::index(double value) const
{
	if (value < _lower || value >= _upper)
		throw NoSuchBin();

	unsigned size = _lowerBounds.size();
	long lo = 0,
		 hi = size - 1,
		 mid = (lo+hi)/2;
	while (lo < hi) {
        if (value < _lowerBounds[mid])
            hi = mid - 1;
        else if (value >= _lowerBounds[mid+1])
            lo = mid + 1;
		else
			break;
        mid = (lo+hi)/2;
    }

	return mid;
} // NonUniformilyBinnedAxis::index()


unsigned &NonUniformilyBinnedAxis::operator[](double value)
{
	return _v[index(value)];
} // NonUniformilyBinnedAxis::operator[]


NonUniformilyBinnedAxis::~NonUniformilyBinnedAxis()
{
	delete [] _v;
	_v = 0;
} // NonUniformilyBinnedAxis::~NonUniformilyBinnedAxis()


template<class NumericType>
LogarithmicGrid<NumericType>::LogarithmicGrid(unsigned gridSize,
											  NumericType maxValue)
{
	_i0 = DBL(gridSize-1)/log(maxValue+1.);
	this->resize(gridSize);

	for (unsigned n=0; n<gridSize; ++n)
		(*this)[n] = value(n);
} // LogarithmicGrid<NumericType>::LogarithmicGrid()


template<class NumericType>
NumericType LogarithmicGrid<NumericType>::value(unsigned n) const
{
	return exp(DBL(n)/_i0)-1.;
} // LogarithmicGrid<NumericType>::value()

//	explicit template instantiation
template class LogarithmicGrid<double>;
