//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalDataVector.cc	--- implementation of CalDataVector
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-05-31: 1.0.0: - first version
//		UL, 1999-12-23: 1.1.0: - added CalIntVector<int32>
//		UL, 2000-03-08: 1.2.0: - added CalVector::clone()
//
//	AUTHOR:
//		U. Lammers, Mon May 31 16:45:25 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CalDataVector.cc,v 1.20 2009/07/06 14:25:19 aibarra Exp $
//

#include <cstring>

#include "errstr.h"
#include "common.h"
#include "Random.h"
#include "CalDataVector.h"
#include "CalServer.h"

using namespace std;

template<class ElementType>
CalVector<ElementType>::CalVector(ElementType *ptr, unsigned n)
	: _ptr(ptr), _n(n), _pcount(0)
{
	if (!ptr && n) {
		_ptr = new ElementType[n];
		_pcount = new unsigned(1);	//	there is now 1 reference
	}
} // CalVector::CalVector()


template<class ElementType>
CalVector<ElementType> &CalVector<ElementType>::operator=(
										const CalVector<ElementType> &rhs)
{
	if (this == &rhs)
		return *this;

	//  check if this is the last reference to object - if yes, delete
	//  it + reference counter
	if (_pcount && --(*_pcount) == 0) {
		delete _ptr;
		delete _pcount;
	}

	_ptr = rhs._ptr;
	_pcount = rhs._pcount;
	if (_pcount)
		++(*_pcount);
	_n = rhs._n;

	return *this;
} // CalVector<ElementType>::operator=()


template<class ElementType>
CalVector<ElementType>::CalVector(const CalVector<ElementType> &src)
	: _ptr(0), _n(0), _pcount(0)
{
	operator=(src);
} // CalVector<ElementType>::CalVector()


template<class ElementType>
void CalVector<ElementType>::clone(const CalVector<ElementType> &source)
{
	if (_ptr == source._ptr)
		return;

	if (_pcount && --(*_pcount) == 0) {
		delete _ptr;
		delete _pcount; _pcount = 0;
	}

	_n				= source._n;
	_ptr 			= new ElementType[_n];
	_pcount			= new unsigned(1);
	memcpy(_ptr, source._ptr, _n * sizeof(ElementType));
} // CalVector::clone()

/* AI gcc4 porting
 * Duplicate explicit instantiation of ‘void CalVector<ElementType>::clone(const CalVector<ElementType>&) [with ElementType = unsigned char]’
 * => Move specialization just after the definition
 */
template void CalVector<int8>::clone(const CalVector<int8> &);
template void CalVector<int16>::clone(const CalVector<int16> &);
template void CalVector<int32>::clone(const CalVector<int32> &);
template void CalVector<real32>::clone(const CalVector<real32> &);
template void CalVector<real64>::clone(const CalVector<real64> &);

template<class ElementType>
CalVector<ElementType>::~CalVector()
{
	if (_pcount && --(*_pcount) == 0) {
		//  this is the last reference - delete object + associated
		//  pointer
		delete _ptr;
		delete _pcount;
	}
} // CalVector<ElementType>::~CalVector()


template<class IntType>
CalRealVector<real32> &CalIntVector<IntType>::toReal(CalRealVector<real32> &rv)
const
{
	/* AI gcc4 porting
	 *  '_n' was not declared in this scope => this->
	 */
	if (rv.size() != this->_n)
		errstr<<"Size of vectors do not match"<<Sas::error("SizeMismatch");


	const CalServer *p = parent();
	if (!p)
		p = &calServer;

	if (p->state()->randomize()->get())
		for (unsigned n=0; n<this->_n; ++n)
			rv[n] = FLT(this->_ptr[n]) + randGen.rand();
	else
		for (unsigned n=0; n<this->_n; ++n)
			rv[n] = FLT(this->_ptr[n]);

	return rv;
} // CalIntVector<IntType>::toReal()


template<class RealType>
template<class ToIntType> CalIntVector<ToIntType>
	&CalRealVector<RealType>::toInt(CalIntVector<ToIntType> &iv) const
{
	if (iv.size() != this->_n)
		errstr<<"Size of vectors do not match"<<Sas::error("SizeMismatch");

	for (unsigned n=0; n<this->_n; ++n)
		iv[n] = this->_ptr[n] >= 0. ? S_CAST(this->_ptr[n] + .5, ToIntType)
							  : S_CAST(this->_ptr[n] - .5, ToIntType);

	return iv;
} // CalIntVector<RealType>::toInt()


//
//	explicit template instantiation
//
template class CalVector<int8>;
template class CalVector<int16>;
template class CalVector<int32>;
template class CalVector<real32>;
template class CalVector<real64>;

template class CalIntVector<int8>;
template class CalIntVector<int16>;
template class CalIntVector<int32>;
template class CalRealVector<real32>;
template class CalRealVector<real64>;

template CalIntVector<int8>
		&CalRealVector<real32>::toInt(CalIntVector<int8> &) const;
template CalIntVector<int16>
		&CalRealVector<real32>::toInt(CalIntVector<int16> &) const;

template CalIntVector<int8>
		&CalRealVector<real64>::toInt(CalIntVector<int8> &) const;
template CalIntVector<int16>
		&CalRealVector<real64>::toInt(CalIntVector<int16> &) const;
template CalIntVector<int32>
		&CalRealVector<real64>::toInt(CalIntVector<int32> &) const;

