//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EncircledEnergyDataServer.cc --- implementation of
//											EncircledEnergyDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the EncircledEnergyDataServer
//		class.
//
//	CHANGE HISTORY:
//		UL, 2001-01-31: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Wed Jan 31 16:17:49 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd, et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/EncircledEnergyDataServer.cc,v 1.8 2009/07/06 14:25:19 aibarra Exp $
//

#include <cmath>
#include <algorithm>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "EncircledEnergyDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( EncircledEnergyFunction )---------------------------------------------
//
EncircledEnergyFunction::EncircledEnergyFunction(double E,
												 double theta,
												 double phi,
					 							 const vector<double> &r,
												 const vector<double> &ee)
	: InterpolatedCurve(r, ee), _E(E), _theta(theta), _phi(phi)
{
} // EncircledEnergyFunction::EncircledEnergyFunction()


//
//	---( EncircledEnergyDataServer )------------------------------------------
//
class MM2ArcSec {
	public:
		MM2ArcSec(double factor) : _factor(factor) {}
		double operator()(double v) { return v*_factor; }
	private:
		double _factor;
};

EncircledEnergyDataServer::EncircledEnergyDataServer(const Table *eeTab,
				  CalAtomCollection<EncircledEnergyDataServer> *parent)
	: CalChild<CalAtomCollection<EncircledEnergyDataServer> >(parent),
	 _eeFunc(0), _E(0.), _theta(0.), _phi(0.)
{
	double conv = eeTab->attribute("CONV")->asReal();	//	arcsec/mm
   _thetaScale._linC =
					CalOalUtility::optionalAttribute(eeTab, "COR1", 0.),
														//	1/arcsec
   _thetaScale._quadC =
				 	CalOalUtility::optionalAttribute(eeTab, "COR2", 0.);
														//	1/arcsec^2
	_thetaScale.setTheta(0.);

	ColumnData *cdE = eeTab->column("ENERGY")->data(),
			   *cdTheta = eeTab->column("THETA")->data(),
			   *cdN = eeTab->column("N")->data();
	real32 *E = cdE->real32Data(),
		   *theta = cdTheta->real32Data();
	int16 *NR = cdN->int16Data();

	unsigned N = eeTab->rows();
	for (unsigned n=0; n<N; ++n) {
		//	read radius value ...
		vector<double> r;
		CalOalUtility::fillVector(eeTab, "RADIUS", n, r);
		r.erase(r.begin()+NR[n], r.end());

		//	... read EE values ...
		vector<double> ee;
		CalOalUtility::fillVector(eeTab, "EE", n, ee);
		ee.erase(ee.begin()+NR[n], ee.end());

		//	... convert radius values from mm->arcsec ...
		double thetaArcSec = theta[n] * RAD_TO_DEG * 3600.;
		_thetaScale.setTheta(thetaArcSec);
		MM2ArcSec mm2ArcSec(_thetaScale(conv));
		transform(r.begin(), r.end(), r.begin(), mm2ArcSec);

		//	... and construct new EE curve
		const InterpolatedCurve *eeCurve = new InterpolatedCurve(r, ee);

		//	register curve
		MIterOuter i = _eeCurves.find(E[n]);
		if (i != _eeCurves.end()) {
			MIterInner j = i->second.find(thetaArcSec);
			if (j != i->second.end())
				errstr<<"Have read encircled energy curve for E/theta="
					  <<E<<"eV/"<<theta<<"arcsec already"
					  <<Sas::error("CcfIntegrityError");
			i->second[thetaArcSec] = eeCurve;
		} else {
			//	new energy value
			MapTypeInner innerMap;
			innerMap[thetaArcSec] = eeCurve;
			_eeCurves[E[n]] = innerMap;
		}
	}
	delete cdE;
	delete cdTheta;
	delete cdN;

	//
	//	establish validity intervals
	//
	_eVal = EncircledEnergy::ValidityInterval(_eeCurves.begin()->first,
											  (--_eeCurves.end())->first);
	_thetaVal = EncircledEnergy::ValidityInterval(0.,
								  (--_eeCurves.begin()->second.end())->first);
	_phiVal = EncircledEnergy::ValidityInterval(0., M_TWO_PI);

	//
	//	we need at least two curves for each energy; check this in the
	//	following and if there is only one on-axis curve register a duplicate
	//	with theta=1800" to avoid problems with the 2-D interpolation
	//
	for (MIterOuter i = _eeCurves.begin(); i != _eeCurves.end(); ++i)
		if (i->second.size() < 2) {
			MapTypeInner &mi = i->second;
			mi[mi.begin()->first + 1800.] =
				new InterpolatedCurve(*mi.begin()->second);
			_thetaVal = EncircledEnergy::ValidityInterval(0., 1800.);
		}
} // EncircledEnergyDataServer::EncircledEnergyDataServer()


void EncircledEnergyDataServer::dump(ostream &)
{
	errstr<<"Sorry dumping EncircledEnergyDataServer not yet implemented"
		  <<Sas::warning("NotYetImplemented");
} // PsfDataServer::dump()


void EncircledEnergyDataServer::setParameters(double E, double theta,
											  double phi)
{
	_E = E;
	if (E<_eVal.lo() || E>_eVal.hi())
		errstr<<"Request encircled energy at energy E="<<E<<" eV which is "
				"outside the validity range ["<<_eVal.lo()<<", "
			  <<_eVal.hi()<<"]"
			  <<Sas::warning("EnergyOutsideValidityRange");

	_theta = theta;
	if (theta<_thetaVal.lo() || theta>_thetaVal.hi())
		errstr<<"Request encircled energy at off-angle theta="<<theta
			  <<" arcsec which is "
				"outside the validity range ["<<_thetaVal.lo()<<", "
			  <<_thetaVal.hi()<<"]"
			  <<Sas::warning("OffaxisAngleOutsideValidityRange");

	_phi = fmod(phi, M_TWO_PI);
	if (_phi < 0.)
		_phi += M_TWO_PI;
	if (_phi<_phiVal.lo() || phi>_phiVal.hi())
		errstr<<"Request encircled energy at azimuth-angle phi="
			  <<phi*RAD_TO_DEG<<" deg "
				"which is outside the validity range ["
			  <<_phiVal.lo()*RAD_TO_DEG<<", "<<_phiVal.hi()*RAD_TO_DEG<<"]"
			  <<Sas::warning("AzimuthAngleOutsideValidityRange");
} // EncircledEnergyDataServer::setParameters()


const EncircledEnergy::ValidityInterval &
	EncircledEnergyDataServer::eValidity() const
{
	return _eVal;
} // EncircledEnergy::ValidityInterval()


const EncircledEnergy::ValidityInterval &
	EncircledEnergyDataServer::thetaValidity() const
{
	return _thetaVal;
} // EncircledEnergyDataServer::thetaValidity()


const EncircledEnergy::ValidityInterval &
	EncircledEnergyDataServer::phiValidity() const
{
	return _phiVal;
} // EncircledEnergyDataServer::phiValidity()


EncircledEnergyFunction &EncircledEnergyDataServer::encircledEnergy(double E,
													 double	theta,
													 double phi)
{
	setParameters(E, theta, phi);

	double thetaLookUp = 0.;
	MIterOuter ior = _eeCurves.upper_bound(E);
	if (ior == _eeCurves.end() || ior == _eeCurves.begin())
		errstr<<"Encircled energy can only be chosen in the interval ["
			  <<_eVal.lo()<<", "<<_eVal.hi()<<"] - requsted energy is: "<<E
			  <<Sas::error("EnergyOutOfValidBounds");
	MIterOuter iol = ior; --iol;
	double E0 = iol->first,
		   E1 = ior->first;

	MIterInner iiur = ior->second.upper_bound(thetaLookUp);
	if (iiur == ior->second.end() || iiur == ior->second.begin())
		errstr<<"Encircled energy off-axis angle can only be chosen in the "
			  <<"interval ["<<_thetaVal.lo()<<", "<<_thetaVal.hi()
			  <<"] - requested angle is: "<<theta
			  <<Sas::error("OffAxisAngleOutOfValidBounds");
	MIterInner iilr = iiur; --iilr;
	double theta0 = iilr->first,
		   theta1 = iiur->first;

	MIterInner iiul = iol->second.upper_bound(thetaLookUp);
	assert(iiul != iol->second.begin() && iiul != iol->second.end());
	MIterInner iill = iiul; --iill;

	double tE = (E-E0)/(E1-E0),
		   ut = (thetaLookUp-theta0)/(theta1-theta0);
	const InterpolatedCurve &eeLL = *iill->second,
						    &eeLR = *iilr->second,
						    &eeUR = *iiur->second,
							&eeUL = *iiul->second;
	//
	//   |    eeUL|        |eeUR
	//	-+--------+--------+--------
	//	 |        |        |             ^
	//   |        |        |        theta|
	//	-+--------+--------+--------     +---> E
	//   |    eeLL|        |eeLR
	//
	vector<double> x(eeLL.abscissa());
	unsigned N = x.size();
	vector<double> y(N);
	for (unsigned n=0; n<N; ++n)
		y[n] = (1.-tE)*(1.-ut) * eeLL(x[n]) +
			   tE*(1.-ut) * eeLR(x[n]) +
			   tE*ut * eeUR(x[n]) +
			   (1.-tE)*ut * eeUL(x[n]);
	_thetaScale.setTheta(theta);
	transform(x.begin(), x.end(), x.begin(), _thetaScale);

	delete _eeFunc;
	return *(_eeFunc = new EncircledEnergyFunction(_E, _theta, _phi, x, y));
} // EncircledEnergyDataServer::encircledEnergy()


EncircledEnergyDataServer::~EncircledEnergyDataServer()
{
	for (MIterOuter i = _eeCurves.begin(); i != _eeCurves.end(); ++i)
		for (MIterInner j = i->second.begin(); j != i->second.end(); ++j)
			delete j->second;
	delete _eeFunc;
} // EncircledEnergyDataServer::~EncircledEnergyDataServer()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<EncircledEnergyDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	EncircledEnergyDataServer *null = 0;
	_collection = new vector<EncircledEnergyDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::XEncirEn);
	DataSet *set = e->open();

	CcfConstituentTable eetab(set, "XENCIREN");
	(*_collection)[0] = new EncircledEnergyDataServer(eetab.block(), this);

	dataSetServer->close(set);
} // CalAtomCollection<EncircledEnergyDataServer>::CalAtomCollection()


/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
EncircledEnergyDataServer *
	CalAtomCollection<EncircledEnergyDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<EncircledEnergyDataServer>::select()
