//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmColorTransformator --- class related to OM color transformation
//
//	DESCRIPTION:
//		Instances of class OmColorTransformator can perform OM color
//		transformations.
//
//	CHANGE HISTORY:
//		UL, 2000-01-11: 1.2.0: - initial version
//
//	AUTHOR:
//		U. Lammers (UL), Tue Jan 11 16:42:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/OmColorTransformator.h,v 1.11 2007/04/26 09:36:05 rsaxton Exp $
//

#ifndef OM_COLOR_TRANSFORMATOR
#define OM_COLOR_TRANSFORMATOR

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalCommon.h"
#include "CalDataVector.h"

#include <vector>

//
//	---( OmColorTransformator )------------------------------------------------
//
class OmColorTransformator : public CalAtom,
		  public virtual CalChild<CalAtomCollection<OmColorTransformator> > {
public:

	OmColorTransformator(const CcfConstituentTable &colorTransTab,
						 CalAtomCollection<OmColorTransformator> *parent = 0);
		//
		//	PURPOSE:
		//		Construct OmColorTransformator object.
		//
		//	ARGUMENTS:
		//		colorTransTab:	table with color transformation data
		//


	static unsigned id() { return Ccf::ColorTrans; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual const OmColorTransformator &withFilters(Filter filter1,
													Filter filter2) const;
		//
		//	PURPOSE:
		//		Define the filters for which the color transformation
		//		is to be carried out
		//
		//	ARGUMENTS:
		//		filter1/2	: filters to select the transformation;
		//					  the first one must be at shorter wavelengths
		//					  than the second one otherwise a fatal
		//					  error condition is raised
		//


	virtual unsigned short branches() const;
		//
		//	PURPOSE:
		//		Inquire the number of available transformation branches
		//		after two filters have been selected with withFilters()
		//		(see above)
		//
		//	RETURNS:
		//		Number of transformation branches; is always >=1
		//


	virtual const OmColorTransformator &branch(unsigned short branchId) const;
		//
		//	PURPOSE:
		//		Select the branch which is to be used for the color
		//		transformation; this is optional - if not called prior
		//		to standardColor() the standard branch 0 (main sequence
		//		stars) shall be used.
		//
		//	ARGUMENTS:
		//		branchId	: numeric identifier in the range [0, max-1]
		//					  where max is the number of available branches
		//					  returned by function branches() above;
		//					  selecting a branch outside the validity interval
		//					  shall lead to an error condition
		//


	typedef Interval<double> ValidityInterval;
	virtual vector<ValidityInterval> validityRanges() const;
		//
		//	PURPOSE:
		//		Inquire the magnitude difference validity ranges for which
		//		the color transformation is valid; standardColor() below
		//		will automatically select the transformation which is
		//		applicable for the given magnitude difference;
		//


	virtual ValueError standardColor(double magnitude1,
									 double magnitude2) const;
		//
		//	PURPOSE:
		//		Compute the color corresponding to two input magnitudes
		//		in the standard system. A prior call to branch() selects
		//		a particular branch of the available color transformations
		//		for the input filters set with withFilters(); calling
		//		branch() this is optional - if not done the default branch 0
		//		(main sequence stars) is used.
		//
		//	ARGUMENTS:
		//		magnitude1/magnitude2:
		//						 The two input magnitudes associated with
		//						 the two filter set via withFilters() above;
		//						 the magnitude difference
		//						    magnitude1 - magnitude2
		//						 must lie in exactly one of the validity range
		//						 reported by valdityRanges() otherwise
		//						 an error condition shall be raised;
		//						 the function will automatically select
		//						 the transformation applicable for the
		//						 given magnitude difference
		//


	virtual ValueError standardMagnitude(ValueError standardColor,
										 double magnitude) const;
		//
		//	PURPOSE
		//		Transform an instrumental magnitude into its standard system;
		//		uses filters defined with withFilters() above
		//
		//	ARGUMENTS:
		//		standardColor  : standard color as computed with
		//						 standardColor() above
		//		magnitude	   : magnitude value to convert
		//		


	virtual ValueAsymmetricError toNaturalMagnitudes(
								const ValueAsymmetricError &naturalRate) const;
		//
		//	PURPOSE:
		//		Converts count rates into magnitudes of the natural system
		//
		//	ARGUMENTS:
		//		naturalRate: corrected count rate + associated errors to be
		//					 converted into natural magnitudes
		//					 


	virtual double apertureRadius() const;
		//
		//	RETURNS:
		//		Aperture radius [pixel] used in photometric calibration.
		//

    struct OmFluxConvF {
        double   cfwd, // WD flux conversion factor
                 cfab, // AB system flux conversion factor
                 zpab, // AB system zero point
               cfvega, // VEGA flux conversion factor
               zpvega; // VEGA system zero point
    };

	virtual OmFluxConvF fluxConversionFactors() const;
		//
		//	RETURNS:
		//		Flux conversion factors for current filter
		//


	virtual ~OmColorTransformator();
		//
		//	PURPOSE:
		//		Destruct OmColorTransformator object
		//

private:
	//	data
	typedef struct {
				std::vector<double>	colorTrafo,
									colorTrafoError,
									magniTrafo,
									magniTrafoError;
			} ColorTrafo;

	typedef std::map<ValidityInterval, ColorTrafo>	ColorTrafoList;
	typedef std::vector<ColorTrafoList>				ColorTrafoBranches;
	typedef std::pair<Filter, Filter>				FilterPair;
	typedef std::map<FilterPair, ColorTrafoBranches> ColorTrafoMap;
	ColorTrafoMap								_colorTrafoMap;

	typedef std::map<Filter, double>			FilterPropertyMap;
	FilterPropertyMap							_zeroPointMap,
                                                _zeroPointAB,
												//  zero points of the
												//	count-rate to magnitude
												//	conversion (AB system)
												_apertureRadiiMap,
												//	aperature radii used in
												//	photometric calibration -
												//	see [2] for details
												_fluxConvWD,
												_fluxConvAB,
												_fluxConvVEGA,
												//	count rate to flux
												//	conversion factors
                                                _zeroPointVEGA;
												//  zero points of the
												//	count-rate to magnitude
												//	conversion (VEGA system)

	mutable Filter		_f1, _f2;
	mutable	unsigned	_branch;

	mutable ColorTrafoMap::const_iterator _selectedTrafoList;
	mutable ColorTrafo					  *_selectedTrafo;

	//	methods
	void registerTrafo(Filter f1, Filter f2, unsigned short branch,
					   const ValidityInterval &vi,
					   const ColorTrafo &colorTrafo);

	void checkFilters() const;

	//	ALGOID=0/1 transformations
	ValueError standardColor0(double instColor) const;
	ValueError standardColor1(double instColor) const;

	ValueError standardMagnitude0(ValueError standardColor,
								  double magnitude) const;
	ValueError standardMagnitude1(ValueError standardColor,
								  double magnitude) const;

	const ColorTrafo &selectedTrafo() const;
};

#endif
