//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RgsBackgroundServer.cc --- implementation of RgsBackgroundServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2004-7-10: 1.0.0: - initial version
//		RDS, 2006-5-12: 1.1.0: - Added XPSF and PIPSF fractions
//
//	AUTHOR:
//		R. Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/RgsBackgroundServer.cc,v 1.5 2010/04/06 11:49:28 rsaxton Exp $
//
#include <iostream>

#include "RgsBackgroundServer.h"
#include "CalOalUtility.h"
#include "stringutil.h"

using namespace std;

//
//	---( RgsBackgroundServer> ) ---------------------------------------
//
RgsBackgroundServer::RgsBackgroundServer(DataSet *backset,
						CalAtomCollection<RgsBackgroundServer> *parent)
		: CalChild<CalAtomCollection<RgsBackgroundServer> >(parent),
          _doubletol(0.0001)
{
    vector<unsigned> levelCount(3);

    _ccfUnits=false;

    unsigned nBlocks = backset->blocks();

    double last_xpsf=0.0, last_pipsf=0.0;

   // Loop over each table
    for (unsigned na=0; na<nBlocks; ++na) {

	    CcfConstituentTable ccfTabSpec(backset, na, 2); // Know about ALGOID=1,2
        const Table *table = ccfTabSpec.block();

      // Get the exposure, order and background level from attributes
        double exposure = table->attribute("EXPOSURE")->asReal();
        unsigned order = table->attribute("ORDER")->asInt();
        double level = table->attribute("LEVEL")->asReal();

      /* If algoid=1 file, get the PSF fraction in the cross-dispersion
         and PI directions. Otherwise set both them to 1.0 and  0.9
         which were the implicit values for the original CCFs */
        double xpsf=1.0, pipsf=0.9;
        unsigned algoid=0;

        if (table->hasAttribute("ALGOID"))
           algoid = table->attribute("ALGOID")->asInt();

        if (algoid>=1) {
          xpsf = table->attribute("XPSFFRA")->asReal();
          pipsf = table->attribute("PIPSFFRA")->asReal();
        }
    
      /* Read the Units string if available */
        string units = "BETA";
        if (algoid>=2) {
           units = table->attribute("UNITS")->asString();
           _ccfUnits=true;
        }

      /* Read in the column data - coded longhand because the
         fillVector is not working */
        vector<int16> channel, quality;
        vector<int32> counts;
        vector<float> areascal, backscal;

      // Get the channel, counts. quality, areascal and backscal arrays
        ColumnData *colchan = table->column("CHANNEL")->data();
        ColumnData *colcounts = table->column("COUNTS")->data();
        ColumnData *colqual = table->column("QUALITY")->data();
        ColumnData *colarea = table->column("AREASCAL")->data();
        ColumnData *colback = table->column("BACKSCAL")->data();
        int16 *channel2 = colchan->int16Data();
        int32 *counts2 = colcounts->int32Data();
        int16 *quality2 = colqual->int16Data();
        real32 *area2 = colarea->real32Data();
        real32 *back2 = colback->real32Data();

        unsigned N = colchan->elements();
        for (unsigned n=0; n<N; ++n) {
            channel.push_back(channel2[n]);
            counts.push_back(counts2[n]);
            quality.push_back(quality2[n]);
            areascal.push_back(area2[n]);
            backscal.push_back(back2[n]);
        }

      // Remove the column data pointers
        delete colchan;
        delete colcounts;
        delete colqual;
        delete colarea;
        delete colback;

        _background.push_back( new RgsBackgnd(exposure, order, level,
                           units, xpsf, pipsf,
                           channel, counts, quality, areascal, backscal) );

      /* Keep a running count of the number of levels available for each order
         Make the assumption that the blocks are grouped together by
         PI and cross-dispersion fraction */
        if (fabs(xpsf - last_xpsf)>_doubletol ||   // New set of values
             fabs(pipsf - last_pipsf)>_doubletol)  // for this xpsf, pipsf
        {
	       for (unsigned i=0; i<levelCount.size(); i++)
              levelCount[i]=0;                     // Zero level indices

           last_xpsf=xpsf;
           last_pipsf=pipsf;
        }

        _orderArray.push_back(order);
        _levelArray.push_back(levelCount[order]);
        _unitsArray.push_back(units);
        _xpsfArray.push_back(xpsf);
        _pipsfArray.push_back(pipsf);
        levelCount[order]++;
    }
   /* Set default value of 100% for the cross-dispersion psf fraction */
    xpsfFrac(1.0);  // Note - this will fail if the CCF is changed but
                    // since the requirements are somewhat fluid ;-)

} // RgsBackgroundServer::RgsBackgroundServer()

void RgsBackgroundServer::dump(ostream &os)
{
	os<<"RgsBackgroundServer{\n";
	os<<"Channels\n"
	  <<"counts\n"<<' '<<endl;
    //vector<unsigned> chn = _background[0]->_channel();
    //vector<int> cnts = _background[0]->_counts();
	//for (unsigned n=0; n<chn.size(); n++)
		//os<<chn[n]<<' '<<cnts[n]<<' '<<endl;
	os<<"}";
} // RgsBackgroundServer::dump()


RgsBackgroundServer::~RgsBackgroundServer()
{
} // RgsBackgroundServer::~RgsBackgroundServer()


/* Method: index() - select the correct background */
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
unsigned RgsBackgroundServer::index() const
{
    assert(_levelCalled);
    assert(_unitsCalled);
    assert(_pipsfCalled);
    assert(_xpsfCalled);

  /* Find position in background array for the current order, xpsf, pipsf,
      index setting */
    for (unsigned i=0; i<_orderArray.size(); i++) {
      if (_orderArray[i]==_orderRequired && _levelArray[i]==_levelRequired &&
          _unitsArray[i]==_unitsRequired &&
          fabs(_xpsfArray[i]-_xpsfRequired)<_doubletol &&
          fabs(_pipsfArray[i]-_pipsfRequired)<_doubletol)
        {
         return i;
        }
    }

    return 0;

} // RgsBackgroundServer::index


/* Method: numberExtensions() - get the number of extns for selected order */
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
unsigned RgsBackgroundServer::numberExtensions() const
{
    if (_orderCalled && _unitsCalled &&_xpsfCalled && _pipsfCalled)
    {
      unsigned count=0;

      for (unsigned i=0; i<_orderArray.size(); i++)
        if (_orderArray[i]==_orderRequired && _xpsfArray[i]==_xpsfRequired
            && _pipsfArray[i]==_pipsfRequired 
                    && _unitsArray[i]==_unitsRequired)
             count++;

      return count;
    }
    else {
      return ( _orderArray.size() ) ;
    }

} // RgsBackgroundServer::numberExtensions

/* Method: order(uint) - set the grating order required */
const RgsBackgroundServer &RgsBackgroundServer::order(unsigned inputOrder)
{
    _orderCalled = true;
    _orderRequired = inputOrder;

    return *this;
} // RgsBackgroundServer::order

/* Method: units(string) - set the units (BETA/LAMBDA) required */
const RgsBackgroundServer &RgsBackgroundServer::units(string inputUnits)
{

    if (inputUnits == "LAMBDA" && !_ccfUnits) 
       errstr<<"LAMBDA not supported by background template CCF "
              <<"- upgrade to the most recent CCF set"
              <<Sas::error("LambdaUnsupported");

    _unitsCalled = true;
    _unitsRequired = inputUnits;

    
    return *this;
} // RgsBackgroundServer::units


/* Method: pipsfFrac(double) - set the PI psf Fraction required */
const RgsBackgroundServer &RgsBackgroundServer::pipsfFrac(double inputPiPsf)
{
   // Ensure that at least one record has this value
    unsigned i;
    for (i=0; i<_orderArray.size(); i++)
      if ( fabs(_pipsfArray[i]-inputPiPsf) < _doubletol ) break;

    if ( i==_orderArray.size() )
       errstr<<"PI PSF fraction "<<inputPiPsf
              <<" not found in background template CCF"
              <<Sas::error("BadPSFFraction");

    _pipsfCalled= true;
    _pipsfRequired = inputPiPsf;

    return *this;
} // RgsBackgroundServer::pipsfFrac

/* Method: xpsfFrac(double) - set the cross-dispersion  required */
const RgsBackgroundServer &RgsBackgroundServer::xpsfFrac(double inputXPsf)
{
   // Ensure that at least one record has this value
    unsigned i;
    for (i=0; i<_orderArray.size(); i++)
      if ( fabs(_xpsfArray[i] - inputXPsf) < _doubletol ) break;

    if ( i==_orderArray.size() )
       errstr<<"Cross-dispersion PSF fraction "<<inputXPsf
              <<" not found in background template CCF"
              <<Sas::error("BadPSFFraction");

    _xpsfCalled= true;
    _xpsfRequired = inputXPsf;

    return *this;
} // RgsBackgroundServer::xpsfFrac

/* Method: level(unsigned) - set the background level required.
                           The level is input as an integer and is simply
                           an index into the array rather than the real level
*/
const RgsBackgroundServer &RgsBackgroundServer::level(unsigned inputLevel)
{
    if (inputLevel >= numberExtensions())
      errstr<<"Background level requested "<<inputLevel
            <<" outside range 0:"<<numberExtensions()-1
            <<Sas::error("BadBckgndLevel");

    _levelCalled = true;
    _levelRequired = inputLevel;

    return *this;
} // RgsBackgroundServer::level

/* Method: exposure() - returns the exposure time for this backgnd map */
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
double RgsBackgroundServer::exposure() const
{
    return _background[index()]->exposure();
} // RgsBackgroundServer::exposure()

/* Method: bckgndLevel() - returns the bckgnd level for this backgnd map */
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
double RgsBackgroundServer::bckgndLevel() const
{
    return _background[index()]->level();
} // RgsBackgroundServer::bckgndLevel()

/* Method: channel() - returns the channel array for this backgnd map */
const vector<int16> RgsBackgroundServer::channel() const
{
    return _background[index()]->channel();
} // RgsBackgroundServer::channel()

/* Method: channel() - returns the counts array for this backgnd map */
const vector<int> RgsBackgroundServer::counts() const
{
    return _background[index()]->counts();
} // RgsBackgroundServer::counts()

/* Method: quality() - returns the quality array for this backgnd map */
const vector<int16> RgsBackgroundServer::quality() const
{
    return _background[index()]->quality();
} // RgsBackgroundServer::quality()

/* Method: areascal() - returns the areascal array for this backgnd map */
const vector<float> RgsBackgroundServer::areascal() const
{
    return _background[index()]->areascal();
} // RgsBackgroundServer::areascal()

/* Method: backscal() - returns the backscal array for this backgnd map */
const vector<float> RgsBackgroundServer::backscal() const
{
    return _background[index()]->backscal();
} // RgsBackgroundServer::backscal()




//
//	---( CalAtomCollection<RgsBackgroundServer> ) ----------------------
//
/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<RgsBackgroundServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	RgsBackgroundServer *null = 0;
	_collection = new vector<RgsBackgroundServer *>(_nEntries, null);
	//
	//	Open the CCF file
	//
	CifEntry *e = ccf->entry(instrument, Ccf::TemplateBckgnd);
	DataSet *dataSet = e->open();

    // Instantiate the server object
	(*_collection)[0] = new RgsBackgroundServer(dataSet, this);


	dataSetServer->close(dataSet);
} // CalAtomCollection<RgsBackgroundServer>::CalAtomCollection()

/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
RgsBackgroundServer *CalAtomCollection<RgsBackgroundServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<RgsBackgroundServer>::select


 //
//	---( RGS background class )---------------------------------------------------------------
//
RgsBackgnd::RgsBackgnd(double exposure, unsigned order, double level,
             string units, double xPsfFrac, double piPsfFrac,
             vector<int16> channel, vector<int> counts,
             vector<int16> quality, vector<float>areascal,
             vector<float>backscal) :
                  _exposure(exposure), _order(order), _level(level),
                  _units(units), _xpsf(xPsfFrac), _pipsf(piPsfFrac),
                  _channel(channel), _counts(counts), _quality(quality),
                  _areascal(areascal), _backscal(backscal)
{
} // Constructor

/* Method: exposure() - returns the exposure time for this backgnd map */
/* AI gcc4 porting
* type qualifiers ignored on function return type
* => remove const
*/
double RgsBackgnd::exposure() const
{
    return _exposure;
} // RgsBackgnd::exposure()
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
double RgsBackgnd::xPsfFrac() const
{
    return _xpsf;
} // RgsBackgnd::xPsfFrac()
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
double RgsBackgnd::piPsfFrac() const
{
    return _pipsf;
} // RgsBackgnd::piPsfFrac()

/* Method: level() - returns the background level for this backgnd map */
/* AI gcc4 porting
 * type qualifiers ignored on function return type
 * => remove const
 */
double RgsBackgnd::level() const
{
    return _level;
} // RgsBackgnd::level()

/* Method: channel() - returns the channel array for this backgnd map */
const vector<int16> RgsBackgnd::channel() const
{
    return _channel;
} // RgsBackgnd::channel()

   /* Method: counts() - returns the counts array for this backgnd map */
const vector<int> RgsBackgnd::counts() const
{
    return _counts;
} // RgsBackgnd::counts()

 /* Method: quality() - returns the quality array for this backgnd map */
const vector<int16> RgsBackgnd::quality() const
{
    return _quality;
} // RgsBackgnd::quality()

 /* Method: areascal() - returns the areascal array for this backgnd map */
const vector<float> RgsBackgnd::areascal() const
{
    return _areascal;
} // RgsBackgnd::areascal()

 /* Method: backscal() - returns the backscal array for this backgnd map */
const vector<float> RgsBackgnd::backscal() const
{
    return _backscal;
} // RgsBackgnd::backscal()

