//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RgsBackgroundServer --- the RgsBackgroundServer class
//
//	DESCRIPTION:
//		A class 'RgsBackgroundServer' is defined which provides access to the
//		RGS template background data
//
//	CHANGE HISTORY:
//		RDS, 2004-07-02: 1.0.0: - first version
//
//	AUTHOR:
//     R.D. Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/RgsBackgroundServer.h,v 1.4 2010/04/06 11:49:28 rsaxton Exp $
//

#ifndef RGS_BACKGROUND_SERVER_H
#define RGS_BACKGROUND_SERVER_H

#include "Dal.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Proxy.h"


//
//	---( RGS background class )---------------------------------------------------------------
//
class RgsBackgnd {
public:

  // Constructor
	RgsBackgnd(double exposure, unsigned order, double level,
             string units, double xPsfFrac, double piPsfFrac,
             vector<int16> channel, vector<int> counts,
             vector<int16> quality, vector<float>areascal,
             vector<float>backscal);
		//
		//	PURPOSE:
		//    Methods to return the data for a background object

	/* AI gcc4 porting
	 * type qualifiers ignored on function return type
	 * => remove const
	 */

	double exposure() const;
	double xPsfFrac() const;
	double piPsfFrac() const;
	double level() const;
  const vector<int16> channel() const;
  const vector<int> counts() const;
  const vector<int16> quality() const;
  const vector<float> areascal() const;
  const vector<float> backscal() const;


 	private:
    double _exposure;
    unsigned _order;
    double _level;
    string _units;
    double _xpsf;
    double _pipsf;
    vector<int16> _channel;
    vector<int> _counts;
    vector<int16> _quality;
    vector<float> _areascal;
    vector<float> _backscal;

}; // RgsBackgnd


//
//	---( RGS background server class )---------------------------------------------------------------
//
class RgsBackgroundServer : public CalAtom,
     public virtual CalChild<CalAtomCollection<RgsBackgroundServer> > {
public:
      RgsBackgroundServer(DataSet *backset,
                          CalAtomCollection<RgsBackgroundServer> *parent = 0);
        //
        //  PURPOSE:
        //      Instantiate RgsBackgroundServer object
        //

          virtual ~RgsBackgroundServer();
        //
        //  PURPOSE:
        //      Destruct RgsBackgroundServer object
        //

          virtual void dump(std::ostream &os);
        //
        //  PURPOSE:
        //      Dump data members to an ostream
        //
        //  ARGUMENTS:
        //      In:
        //          os: stream to print on
        //

          static unsigned id() { return Ccf::TemplateBckgnd; }
        //
        //  PURPOSE:
        //      Return ID of CalAtom
        //



    // Set the required order and level
      const RgsBackgroundServer &order(unsigned inputOrder);
      const RgsBackgroundServer &units(string inputUnits);
      const RgsBackgroundServer &level(unsigned inputLevel);
      const RgsBackgroundServer &pipsfFrac(double inputPiPsf);
      const RgsBackgroundServer &xpsfFrac(double inputXPsf);

    // Return the number of extensions for this order
      /* AI gcc4 porting
       * type qualifiers ignored on function return type
       * => remove const
       */
		  unsigned numberExtensions() const;
		  // Return components of the RgsBackgnd object
		  double exposure() const;
		  double bckgndLevel() const;

      const vector<int16> channel() const;
      const vector<int> counts() const;
      const vector<int16> quality() const;
      const vector<float> areascal() const;
      const vector<float> backscal() const;

protected:
     // Calculate the position of the relevant RgsBackgnd object
		/* AI gcc4 porting
      * type qualifiers ignored on function return type
      * => remove const
      */
	unsigned index() const;

 	private:
    vector<RgsBackgnd *> _background;

    vector<unsigned> _orderArray;
    vector<string> _unitsArray;
    vector<unsigned> _levelArray;
    vector<double> _xpsfArray;
    vector<double> _pipsfArray;

    bool _levelCalled;
    bool _orderCalled;
    bool _unitsCalled;
    bool _xpsfCalled;
    bool _pipsfCalled;
    unsigned  _levelRequired;
    unsigned _orderRequired;
    string _unitsRequired;
    double _pipsfRequired;
    double _xpsfRequired;
    bool _ccfUnits;

   // Tolerance for PSF fraction test
    double _doubletol;


}; // RgsBackgndServer



#endif
