//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeJumpTolServer.cc --- implementation of TimeJumpTolServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the TimeJumpTolServer class
//
//	CHANGE HISTORY:
//		RDS, 2012-12-04: 1.0.0 - first version from CalAtomsCommon
//
//	AUTHOR:
//		Richard Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/TimeJumpTolServer.cc,v 1.2 2013/02/15 10:23:03 rsaxton Exp $
//

#include <cmath>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "TimeJumpTolServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( TimeJumpTolServer )--------------------------------------------------
//
TimeJumpTolServer::TimeJumpTolServer(vector<int> v, vector<int> rev, CalAtomCollection<TimeJumpTolServer> *parent)
	: CalChild<CalAtomCollection<TimeJumpTolServer> >(parent)
{
_v=v;
_rev=rev;
} // TimeJumpTolServer::TimeJumpTolServer()

//
// Method: getJumpTolerance - to return the jump tolerance factor for
//                            this revolution and observing mode
int TimeJumpTolServer::getJumpTolerance(int rev) const
{
   // Handle the extremes
    if (rev <= _rev[0])
       return _v[0];
    else if (rev >= _rev[_v.size()-1])
       return _v[_v.size()-1];

   // Interpolate to find the correct tolerance
    int jumptol=0;
    for (unsigned n=0; n<_v.size()-1; ++n) {
        if (rev > _rev[n] && rev <= _rev[n+1])
           jumptol = INT((_v[n] + (_v[n+1]-_v[n]) * (rev - _rev[n]) /
                                        float((_rev[n+1] - _rev[n]))));
    }
    return jumptol;
}

TimeJumpTolServer::~TimeJumpTolServer()
{
} // TimeJumpTolServer::~TimeJumpTolServer()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<TimeJumpTolServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
        _nModes         = instrument->nModes();
        _nEntries       = _nModes;

	TimeJumpTolServer *null = 0;
	_collection = new vector<TimeJumpTolServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::TimeJumpTol);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, e->queryBlockName(Ccf::toString(Ccf::TimeJumpTol)));

       // Read in the revolution column
	const Table *table = ccftab.block();
        vector<int> rev;
        CalOalUtility::fillVector(table, "REV", rev);

       // Define a tolerance vector for each mode
        vector<int> v;

       //FF
        CalOalUtility::fillVector(table, "TOL_FF", v);
        unsigned mode=xmmPayLoad.epn.modeIndex(Xmm::toMode("PrimeFullWindow"));
        (*_collection)[mode] = new TimeJumpTolServer(v, rev, this);
       //eFF
        CalOalUtility::fillVector(table, "TOL_eFF", v);
        mode=xmmPayLoad.epn.modeIndex(Xmm::toMode("PrimeFullWindowExtended"));
        (*_collection)[mode] = new TimeJumpTolServer(v, rev, this);
       //LW
        CalOalUtility::fillVector(table, "TOL_LW", v);
        mode=xmmPayLoad.epn.modeIndex(Xmm::toMode("PrimeLargeWindow"));
        (*_collection)[mode] = new TimeJumpTolServer(v, rev, this);
       //SW
        CalOalUtility::fillVector(table, "TOL_SW", v);
        mode=xmmPayLoad.epn.modeIndex(Xmm::toMode("PrimeSmallWindow"));
        (*_collection)[mode] = new TimeJumpTolServer(v, rev, this);
       //TI
        CalOalUtility::fillVector(table, "TOL_TI", v);
        mode=xmmPayLoad.epn.modeIndex(Xmm::toMode("FastTiming"));
        (*_collection)[mode] = new TimeJumpTolServer(v, rev, this);
       //BU
        CalOalUtility::fillVector(table, "TOL_BU", v);
        mode=xmmPayLoad.epn.modeIndex(Xmm::toMode("FastBurst"));
        (*_collection)[mode] = new TimeJumpTolServer(v, rev, this);

	dataSetServer->close(dataSet);
} // CalAtomCollection<TimeJumpTolServer>::CalAtomCollection()

template<>
TimeJumpTolServer *CalAtomCollection<TimeJumpTolServer>::select(
                                        Ccd, Node,
                                        unsigned modeOffset) const
{
        if (modeOffset > xmmPayLoad.epn.nModes() || !(*_collection)[modeOffset])
                //      not found
                errstr<<"There is no jump tolerance for EPN mode `"
                          <<Xmm::toString(xmmPayLoad.epn.modes()[modeOffset])<<"'"
                          <<Sas::error("IllegalInput");

        return (*_collection)[modeOffset];
} // TimeJumpTolServer::select

